import copy
import os
import subprocess
import sys

import numpy as np
from scipy.fft import fft, ifft, fftshift
from scipy.io.wavfile import write, read
from scipy.signal import resample
from scipy.signal.windows import blackmanharris, triang


# sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'utilFunctions_C'))

# try:
#     import utilFunctions_C as UF_C
# except ImportError:
#     print("\n")
#     print("-------------------------------------------------------------------------------")
#     print("Warning:")
#     print("Cython modules for some of the core functions were not imported.")
#     print("Please refer to the README.md file in the 'sms-tools' directory,")
#     print("for the instructions to compile the cython modules.")
#     print("Exiting the code!!")
#     print("-------------------------------------------------------------------------------")
#     print("\n")
#     sys.exit(0)


winsound_imported = False
if sys.platform == "win32":
    try:
        import winsound

        winsound_imported = True
    except:
        print("You won't be able to play sounds, winsound could not be imported")


def isPower2(num):
    """
	Check if num is power of two
	"""
    return ((num & (num - 1)) == 0) and num > 0


INT16_FAC = (2 ** 15)
INT32_FAC = (2 ** 31)
INT64_FAC = (2 ** 63)
norm_fact = {'int16': INT16_FAC, 'int32': INT32_FAC, 'int64': INT64_FAC, 'float32': 1.0, 'float64': 1.0}


def wavread(filename):
    """
	Read a sound file and convert it to a normalized floating point array
	filename: name of file to read
	returns fs: sampling rate of file, x: floating point array
	"""

    if (os.path.isfile(filename) == False):  # raise error if wrong input file
        raise ValueError("Input file is wrong")

    fs, x = read(filename)

    if (len(x.shape) != 1):  # raise error if more than one channel
        raise ValueError("Audio file should be mono")

    if (fs != 44100):  # raise error if more than one channel
        raise ValueError("Sampling rate of input sound should be 44100")

    # scale down and convert audio into floating point number in range of -1 to 1
    x = np.float32(x) / norm_fact[x.dtype.name]
    return fs, x


def wavplay(filename):
    """
	Play a wav audio file from system using OS calls
	filename: name of file to read
	"""
    if (os.path.isfile(filename) == False):  # raise error if wrong input file
        print("Input file does not exist. Make sure you computed the analysis/synthesis")
    else:
        if sys.platform == "linux" or sys.platform == "linux2":
            # linux
            subprocess.call(["aplay", filename])

        elif sys.platform == "darwin":
            # OS X
            subprocess.call(["afplay", filename])
        elif sys.platform == "win32":
            if winsound_imported:
                winsound.PlaySound(filename, winsound.SND_FILENAME)
            else:
                print("Cannot play sound, winsound could not be imported")
        else:
            print("Platform not recognized")


def wavwrite(y, fs, filename):
    """
	Write a sound file from an array with the sound and the sampling rate
	y: floating point array of one dimension, fs: sampling rate
	filename: name of file to create
	"""

    x = copy.deepcopy(y)  # copy array
    x *= INT16_FAC  # scaling floating point -1 to 1 range signal to int16 range
    x = np.int16(x)  # converting to int16 type
    write(filename, fs, x)


def peakDetection(mX, t):
    """
	Detect spectral peak locations
	mX: magnitude spectrum, t: threshold
	returns ploc: peak locations
	"""

    thresh = np.where(np.greater(mX[1:-1], t), mX[1:-1], 0)  # locations above threshold
    next_minor = np.where(mX[1:-1] > mX[2:], mX[1:-1], 0)  # locations higher than the next one
    prev_minor = np.where(mX[1:-1] > mX[:-2], mX[1:-1], 0)  # locations higher than the previous one
    ploc = thresh * next_minor * prev_minor  # locations fulfilling the three criteria
    ploc = ploc.nonzero()[0] + 1  # add 1 to compensate for previous steps
    return ploc


def peakInterp(mX, pX, ploc):
    """
	Interpolate peak values using parabolic interpolation
	mX, pX: magnitude and phase spectrum, ploc: locations of peaks
	returns iploc, ipmag, ipphase: interpolated peak location, magnitude and phase values
	"""

    val = mX[ploc]  # magnitude of peak bin
    lval = mX[ploc - 1]  # magnitude of bin at left
    rval = mX[ploc + 1]  # magnitude of bin at right
    iploc = ploc + 0.5 * (lval - rval) / (lval - 2 * val + rval)  # center of parabola
    ipmag = val - 0.25 * (lval - rval) * (iploc - ploc)  # magnitude of peaks
    ipphase = np.interp(iploc, np.arange(0, pX.size), pX)  # phase of peaks by linear interpolation
    return iploc, ipmag, ipphase


def sinc(x, N):
    """
	Generate the main lobe of a sinc function (Dirichlet kernel)
	x: array of indexes to compute; N: size of FFT to simulate
	returns y: samples of the main lobe of a sinc function
	"""

    y = np.sin(N * x / 2) / np.sin(x / 2)  # compute the sinc function
    y[np.isnan(y)] = N  # avoid NaN if x == 0
    return y


def genBhLobe(x):
    """
	Generate the main lobe of a Blackman-Harris window
	x: bin positions to compute (real values)
	returns y: main lobe os spectrum of a Blackman-Harris window
	"""

    N = 512  # size of fft to use
    f = x * np.pi * 2 / N  # frequency sampling
    df = 2 * np.pi / N
    y = np.zeros(x.size)  # initialize window
    consts = [0.35875, 0.48829, 0.14128, 0.01168]  # window constants
    for m in range(0, 4):  # iterate over the four sincs to sum
        y += consts[m] / 2 * (sinc(f - df * m, N) + sinc(f + df * m, N))  # sum of scaled sinc functions
    y = y / N / consts[0]  # normalize
    return y



def genSpecSines(ipfreq, ipmag, ipphase, N, fs):
    """
	Generate a spectrum from a series of sine values, calling a C function
	ipfreq, ipmag, ipphase: sine peaks frequencies, magnitudes and phases
	N: size of the complex spectrum to generate; fs: sampling frequency
	returns Y: generated complex spectrum of sines
	"""

    #Y = UF_C.genSpecSines(N * ipfreq / float(fs), ipmag, ipphase, N)
    return Y


def genSpecSines_p(ipfreq, ipmag, ipphase, N, fs):
    """
	Generate a spectrum from a series of sine values
	iploc, ipmag, ipphase: sine peaks locations, magnitudes and phases
	N: size of the complex spectrum to generate; fs: sampling rate
	returns Y: generated complex spectrum of sines
	"""

    Y = np.zeros(N, dtype=complex)  # initialize output complex spectrum
    hN = N // 2  # size of positive freq. spectrum
    for i in range(0, ipfreq.size):  # generate all sine spectral lobes
        loc = N * ipfreq[i] / fs  # it should be in range ]0,hN-1[
        if loc == 0 or loc > hN - 1: continue
        binremainder = round(loc) - loc
        lb = np.arange(binremainder - 4, binremainder + 5)  # main lobe (real value) bins to read
        lmag = genBhLobe(lb) * 10 ** (ipmag[i] / 20)  # lobe magnitudes of the complex exponential
        b = np.arange(round(loc) - 4, round(loc) + 5, dtype='int')
        for m in range(0, 9):
            if b[m] < 0:  # peak lobe crosses DC bin
                Y[-b[m]] += lmag[m] * np.exp(-1j * ipphase[i])
            elif b[m] > hN:  # peak lobe croses Nyquist bin
                Y[b[m]] += lmag[m] * np.exp(-1j * ipphase[i])
            elif b[m] == 0 or b[m] == hN:  # peak lobe in the limits of the spectrum
                Y[b[m]] += lmag[m] * np.exp(1j * ipphase[i]) + lmag[m] * np.exp(-1j * ipphase[i])
            else:  # peak lobe in positive freq. range
                Y[b[m]] += lmag[m] * np.exp(1j * ipphase[i])
        Y[hN + 1:] = Y[hN - 1:0:-1].conjugate()  # fill the negative part of the spectrum
    return Y


def sinewaveSynth(freqs, amp, H, fs):
    """
	Synthesis of one sinusoid with time-varying frequency
	freqs, amps: array of frequencies and amplitudes of sinusoids
	H: hop size, fs: sampling rate
	returns y: output array sound
	"""

    t = np.arange(H) / float(fs)  # time array
    lastphase = 0  # initialize synthesis phase
    lastfreq = freqs[0]  # initialize synthesis frequency
    y = np.array([])  # initialize output array
    for l in range(freqs.size):  # iterate over all frames
        if (lastfreq == 0) & (freqs[l] == 0):  # if 0 freq add zeros
            A = np.zeros(H)
            freq = np.zeros(H)
        elif (lastfreq == 0) & (freqs[l] > 0):  # if starting freq ramp up the amplitude
            A = np.arange(0, amp, amp / H)
            freq = np.ones(H) * freqs[l]
        elif (lastfreq > 0) & (freqs[l] > 0):  # if freqs in boundaries use both
            A = np.ones(H) * amp
            if (lastfreq == freqs[l]):
                freq = np.ones(H) * lastfreq
            else:
                freq = np.arange(lastfreq, freqs[l], (freqs[l] - lastfreq) / H)
        elif (lastfreq > 0) & (freqs[l] == 0):  # if ending freq ramp down the amplitude
            A = np.arange(amp, 0, -amp / H)
            freq = np.ones(H) * lastfreq
        phase = 2 * np.pi * freq * t + lastphase  # generate phase values
        yh = A * np.cos(phase)  # compute sine for one frame
        lastfreq = freqs[l]  # save frequency for phase propagation
        lastphase = np.remainder(phase[H - 1], 2 * np.pi)  # save phase to be use for next frame
        y = np.append(y, yh)  # append frame to previous one
    return y


def cleaningTrack(track, minTrackLength=3):
    """
	Delete fragments of one single track smaller than minTrackLength
	track: array of values; minTrackLength: minimum duration of tracks in number of frames
	returns cleanTrack: array of clean values
	"""

    nFrames = track.size  # number of frames
    cleanTrack = np.copy(track)  # copy array
    trackBegs = np.nonzero((track[:nFrames - 1] <= 0)  # beginning of track contours
                           & (track[1:] > 0))[0] + 1
    if track[0] > 0:
        trackBegs = np.insert(trackBegs, 0, 0)
    trackEnds = np.nonzero((track[:nFrames - 1] > 0) & (track[1:] <= 0))[0] + 1
    if track[nFrames - 1] > 0:
        trackEnds = np.append(trackEnds, nFrames - 1)
    trackLengths = 1 + trackEnds - trackBegs  # lengths of trach contours
    for i, j in zip(trackBegs, trackLengths):  # delete short track contours
        if j <= minTrackLength:
            cleanTrack[i:i + j] = 0
    return cleanTrack

'''
def f0Twm(pfreq, pmag, ef0max, minf0, maxf0, f0t=0):
    """
	Function that wraps the f0 detection function TWM, selecting the possible f0 candidates
	and calling the function TWM with them
	pfreq, pmag: peak frequencies and magnitudes,
	ef0max: maximum error allowed, minf0, maxf0: minimum  and maximum f0
	f0t: f0 of previous frame if stable
	returns f0: fundamental frequency in Hz
	"""
    if (minf0 < 0):  # raise exception if minf0 is smaller than 0
        raise ValueError("Minimum fundamental frequency (minf0) smaller than 0")

    if (maxf0 >= 10000):  # raise exception if maxf0 is bigger than 10000Hz
        raise ValueError("Maximum fundamental frequency (maxf0) bigger than 10000Hz")

    if (pfreq.size < 3) & (f0t == 0):  # return 0 if less than 3 peaks and not previous f0
        return 0

    f0c = np.argwhere((pfreq > minf0) & (pfreq < maxf0))[:, 0]  # use only peaks within given range
    if (f0c.size == 0):  # return 0 if no peaks within range
        return 0
    f0cf = pfreq[f0c]  # frequencies of peak candidates
    f0cm = pmag[f0c]  # magnitude of peak candidates

    if f0t > 0:  # if stable f0 in previous frame
        shortlist = np.argwhere(np.abs(f0cf - f0t) < f0t / 2.0)[:, 0]  # use only peaks close to it
        maxc = np.argmax(f0cm)
        maxcfd = f0cf[maxc] % f0t
        if maxcfd > f0t / 2:
            maxcfd = f0t - maxcfd
        if (maxc not in shortlist) and (maxcfd > (f0t / 4)):  # or the maximum magnitude peak is not a harmonic
            shortlist = np.append(maxc, shortlist)
        f0cf = f0cf[shortlist]  # frequencies of candidates

    if (f0cf.size == 0):  # return 0 if no peak candidates
        return 0

    f0, f0error = UF_C.twm(pfreq, pmag, f0cf)  # call the TWM function with peak candidates, cython version
    #	f0, f0error = TWM_p(pfreq, pmag, f0cf)        # call the TWM function with peak candidates, python version

    if (f0 > 0) and (f0error < ef0max):  # accept and return f0 if below max error allowed
        return f0
    else:
        return 0

'''
def TWM_p(pfreq, pmag, f0c):
    """
	Two-way mismatch algorithm for f0 detection (by Beauchamp&Maher)
	[better to use the C version of this function: UF_C.twm]
	pfreq, pmag: peak frequencies in Hz and magnitudes,
	f0c: frequencies of f0 candidates
	returns f0, f0Error: fundamental frequency detected and its error
	"""

    p = 0.5  # weighting by frequency value
    q = 1.4  # weighting related to magnitude of peaks
    r = 0.5  # scaling related to magnitude of peaks
    rho = 0.33  # weighting of MP error
    Amax = max(pmag)  # maximum peak magnitude
    maxnpeaks = 10  # maximum number of peaks used
    harmonic = np.matrix(f0c)
    ErrorPM = np.zeros(harmonic.size)  # initialize PM errors
    MaxNPM = min(maxnpeaks, pfreq.size)
    for i in range(0, MaxNPM):  # predicted to measured mismatch error
        difmatrixPM = harmonic.T * np.ones(pfreq.size)
        difmatrixPM = abs(difmatrixPM - np.ones((harmonic.size, 1)) * pfreq)
        FreqDistance = np.amin(difmatrixPM, axis=1)  # minimum along rows
        peakloc = np.argmin(difmatrixPM, axis=1)
        Ponddif = np.array(FreqDistance) * (np.array(harmonic.T) ** (-p))
        PeakMag = pmag[peakloc]
        MagFactor = 10 ** ((PeakMag - Amax) / 20)
        ErrorPM = ErrorPM + (Ponddif + MagFactor * (q * Ponddif - r)).T
        harmonic = harmonic + f0c

    ErrorMP = np.zeros(harmonic.size)  # initialize MP errors
    MaxNMP = min(maxnpeaks, pfreq.size)
    for i in range(0, f0c.size):  # measured to predicted mismatch error
        nharm = np.round(pfreq[:MaxNMP] / f0c[i])
        nharm = (nharm >= 1) * nharm + (nharm < 1)
        FreqDistance = abs(pfreq[:MaxNMP] - nharm * f0c[i])
        Ponddif = FreqDistance * (pfreq[:MaxNMP] ** (-p))
        PeakMag = pmag[:MaxNMP]
        MagFactor = 10 ** ((PeakMag - Amax) / 20)
        ErrorMP[i] = sum(MagFactor * (Ponddif + MagFactor * (q * Ponddif - r)))

    Error = (ErrorPM[0] / MaxNPM) + (rho * ErrorMP / MaxNMP)  # total error
    f0index = np.argmin(Error)  # get the smallest error
    f0 = f0c[f0index]  # f0 with the smallest error

    return f0, Error[f0index]


def sineSubtraction(x, N, H, sfreq, smag, sphase, fs):
    """
	Subtract sinusoids from a sound
	x: input sound, N: fft-size, H: hop-size
	sfreq, smag, sphase: sinusoidal frequencies, magnitudes and phases
	returns xr: residual sound
	"""

    hN = N // 2  # half of fft size
    x = np.append(np.zeros(hN), x)  # add zeros at beginning to center first window at sample 0
    x = np.append(x, np.zeros(hN))  # add zeros at the end to analyze last sample
    bh = blackmanharris(N)  # blackman harris window
    w = bh / sum(bh)  # normalize window
    sw = np.zeros(N)  # initialize synthesis window
    sw[hN - H:hN + H] = triang(2 * H) / w[hN - H:hN + H]  # synthesis window
    L = sfreq.shape[0]  # number of frames, this works if no sines
    xr = np.zeros(x.size)  # initialize output array
    pin = 0
    for l in range(L):
        xw = x[pin:pin + N] * w  # window the input sound
        X = fft(fftshift(xw))  # compute FFT
        #Yh = UF_C.genSpecSines(N * sfreq[l, :] / fs, smag[l, :], sphase[l, :], N)  # generate spec sines, cython version
        Yh = genSpecSines_p(N*sfreq[l,:]/fs, smag[l,:], sphase[l,:], N, fs)   # generate spec sines, python version
        Xr = X - Yh  # subtract sines from original spectrum
        xrw = np.real(fftshift(ifft(Xr)))  # inverse FFT
        xr[pin:pin + N] += xrw * sw  # overlap-add
        pin += H  # advance sound pointer
    xr = np.delete(xr, range(hN))  # delete half of first window which was added in stftAnal
    xr = np.delete(xr, range(xr.size - hN, xr.size))  # delete half of last window which was added in stftAnal
    return xr


def stochasticResidualAnal(x, N, H, sfreq, smag, sphase, fs, stocf):
    """
	Subtract sinusoids from a sound and approximate the residual with an envelope
	x: input sound, N: fft size, H: hop-size
	sfreq, smag, sphase: sinusoidal frequencies, magnitudes and phases
	fs: sampling rate; stocf: stochastic factor, used in the approximation
	returns stocEnv: stochastic approximation of residual
	"""

    hN = N // 2  # half of fft size
    x = np.append(np.zeros(hN), x)  # add zeros at beginning to center first window at sample 0
    x = np.append(x, np.zeros(hN))  # add zeros at the end to analyze last sample
    bh = blackmanharris(N)  # synthesis window
    w = bh / sum(bh)  # normalize synthesis window
    L = sfreq.shape[0]  # number of frames, this works if no sines
    pin = 0
    for l in range(L):
        xw = x[pin:pin + N] * w  # window the input sound
        X = fft(fftshift(xw))  # compute FFT
        #Yh = UF_C.genSpecSines(N * sfreq[l, :] / fs, smag[l, :], sphase[l, :], N)  # generate spec sines, cython version
        Yh = genSpecSines_p(N*sfreq[l,:]/fs, smag[l,:], sphase[l,:], N, fs)   # generate spec sines, python version
        Xr = X - Yh  # subtract sines from original spectrum
        mXr = 20 * np.log10(abs(Xr[:hN]))  # magnitude spectrum of residual
        mXrenv = resample(np.maximum(-200, mXr), mXr.size * stocf)  # decimate the mag spectrum
        if l == 0:  # if first frame
            stocEnv = np.array([mXrenv])
        else:  # rest of frames
            stocEnv = np.vstack((stocEnv, np.array([mXrenv])))
        pin += H  # advance sound pointer
    return stocEnv



import sys
import os
import glob
import numpy as np
import soundfile as sf

import torch
import torch.nn as nn
from torch.nn import functional as F

import librosa
import librosa.display

import matplotlib.pyplot as plt
from matplotlib import collections as mc
from .utils import * 

def plot_waveform(wav, sr, filename=None, title=''):
    librosa.display.waveshow(wav, sr=sr)
    plt.title(title)
    plt.xlabel("time")
    plt.ylabel("amplitude")
    plt.ylim([-1.0, 1.0])
    if filename:
        plt.savefig(filename)
        plt.close()
         
def plot_spec(spec, sr, hop_length, filename=None, title=''):
    librosa.display.specshow(
        spec, 
        y_axis='linear', 
        x_axis='time', 
        sr=sr, 
        hop_length=hop_length,
        cmap='magma')
    plt.title(title)
    if filename:
        plt.savefig(filename)
        plt.close()

def plot_spec_diff(path_y_anno, path_y_pred, path_outdir, sr, num_channel):

    os.makedirs(path_outdir, exist_ok=True)

    # load
    y_anno, sr_anno = sf.read(path_y_anno)
    y_pred, sr_pred = sf.read(path_y_pred)

    assert sr_anno == sr 
    assert sr_pred == sr 

    if num_channel == 1:
        y_anno = y_anno[..., None]
        y_pred = y_pred[..., None]

    # length alignment
    min_len = min(len(y_anno), len(y_pred))
    y_anno = y_anno[:min_len]
    y_pred = y_pred[:min_len]

    # plot spec_diff
    for cidx in range(num_channel):
        FREQ_LIM =(20, sr//2)
        f_axis = np.linspace(0.0, sr/2, 4096//2 + 1)
        plt.figure(figsize=(10, 5))
        plt.xscale('log')

        path_song_spec_diff = os.path.join(path_outdir, f'spec_diff_ch-{cidx}.png')
        y_anno_t = torch.from_numpy(y_anno[:, cidx])
        y_pred_t = torch.from_numpy(y_pred[:, cidx])
        spec_diff = get_smoothed_spectrum_diff(y_pred_t, y_anno_t)
        
        plt.plot(f_axis, spec_diff)

        plt.grid(visible=True, which='major', color='#515A62', linestyle='-')
        plt.grid(visible=True, which='minor', color='grey', linestyle=':')
        plt.xlim(FREQ_LIM)
        plt.minorticks_on()
        plt.xlabel('Frequency', fontsize=14)
        plt.ylabel('Magnitude (dB)', fontsize=14)
        plt.title('Spectrum Difference')
        plt.tight_layout()
        plt.savefig(path_song_spec_diff)
        plt.close()

def plot_wav_displayment(path_y_anno, path_y_pred, path_outdir, sr, num_channel):
    os.makedirs(path_outdir, exist_ok=True)

    # load
    y_anno, sr_anno = sf.read(path_y_anno)
    y_pred, sr_pred = sf.read(path_y_pred)

    assert sr_anno == sr 
    assert sr_pred == sr 

    
    if num_channel == 1:
        y_anno = y_anno[..., None]
        y_pred = y_pred[..., None]

    # length alignment
    min_len = min(len(y_anno), len(y_pred))
    y_anno = y_anno[:min_len]
    y_pred = y_pred[:min_len]

    # start = 20
    # y_anno = y_anno[sr_anno*(start):sr_anno*(start+1)]
    # y_pred = y_pred[sr_anno*(start):sr_pred*(start+1)]
    
    # plot wav_diff
    for cidx in range(num_channel):
        path_song_wav_diff = os.path.join(path_outdir, f'wav_ch-{cidx}.png')

        librosa.display.waveshow(y_anno[:, cidx], sr=sr, label='Target')
        librosa.display.waveshow(y_pred[:, cidx], sr=sr, alpha = 0.5, label='Predict')
        plt.title('Wav diff')
        plt.xlabel("time")
        plt.ylabel("amplitude")
        plt.ylim([-1.0, 1.0])
        plt.tight_layout()
        plt.legend()
        plt.savefig(path_song_wav_diff)
        plt.close()








import numpy as np


def center_crop(x, length):
    start = (x.shape[-1] - length) // 2
    stop = start + length
    return x[..., start:stop]

def causal_crop(x, length):
    stop = x.shape[-1] - 1
    start = stop - length
    return x[..., start:stop]




# functions that implement analysis and synthesis of sounds using the Discrete Fourier Transform
# (for example usage check dftModel_function.py in the models_interface directory)

import numpy as np
import math
from scipy.fft import fft, ifft
from . import utilFunctions as UF
#import utilFunctions as UF

tol = 1e-14  # threshold used to compute phase


def dftModel(x, w, N):
    """
	Analysis/synthesis of a signal using the discrete Fourier transform
	x: input signal, w: analysis window, N: FFT size
	returns y: output signal
	"""

    if not (UF.isPower2(N)):  # raise error if N not a power of two
        raise ValueError("FFT size (N) is not a power of 2")

    if (w.size > N):  # raise error if window size bigger than fft size
        raise ValueError("Window size (M) is bigger than FFT size")

    if all(x == 0):  # if input array is zeros return empty output
        return np.zeros(x.size)
    hN = (N // 2) + 1  # size of positive spectrum, it includes sample 0
    hM1 = (w.size + 1) // 2  # half analysis window size by rounding
    hM2 = int(math.floor(w.size / 2))  # half analysis window size by floor
    fftbuffer = np.zeros(N)  # initialize buffer for FFT
    y = np.zeros(x.size)  # initialize output array
    # ----analysis--------
    xw = x * w  # window the input sound
    fftbuffer[:hM1] = xw[hM2:]  # zero-phase window in fftbuffer
    fftbuffer[-hM2:] = xw[:hM2]
    X = fft(fftbuffer)  # compute FFT
    absX = abs(X[:hN])  # compute ansolute value of positive side
    absX[absX < np.finfo(float).eps] = np.finfo(float).eps  # if zeros add epsilon to handle log
    mX = 20 * np.log10(absX)  # magnitude spectrum of positive frequencies in dB
    pX = np.unwrap(np.angle(X[:hN]))  # unwrapped phase spectrum of positive frequencies
    # -----synthesis-----
    Y = np.zeros(N, dtype=complex)  # clean output spectrum
    Y[:hN] = 10 ** (mX / 20) * np.exp(1j * pX)  # generate positive frequencies
    Y[hN:] = 10 ** (mX[-2:0:-1] / 20) * np.exp(-1j * pX[-2:0:-1])  # generate negative frequencies
    fftbuffer = np.real(ifft(Y))  # compute inverse FFT
    y[:hM2] = fftbuffer[-hM2:]  # undo zero-phase window
    y[hM2:] = fftbuffer[:hM1]
    return y


def dftAnal(x, w, N):
    """
	Analysis of a signal using the discrete Fourier transform
	x: input signal, w: analysis window, N: FFT size 
	returns mX, pX: magnitude and phase spectrum
	"""

    if not (UF.isPower2(N)):  # raise error if N not a power of two
        raise ValueError("FFT size (N) is not a power of 2")

    if w.size > N:  # raise error if window size bigger than fft size
        raise ValueError("Window size (M) is bigger than FFT size")

    hN = (N // 2) + 1  # size of positive spectrum, it includes sample 0
    hM1 = (w.size + 1) // 2  # half analysis window size by rounding
    hM2 = w.size // 2  # half analysis window size by floor
    fftbuffer = np.zeros(N)  # initialize buffer for FFT
    w = w / sum(w)  # normalize analysis window
    xw = x * w  # window the input sound
    fftbuffer[:hM1] = xw[hM2:]  # zero-phase window in fftbuffer
    fftbuffer[-hM2:] = xw[:hM2]
    X = fft(fftbuffer)  # compute FFT
    absX = abs(X[:hN])  # compute ansolute value of positive side
    absX[absX < np.finfo(float).eps] = np.finfo(float).eps  # if zeros add epsilon to handle log
    mX = 20 * np.log10(absX)  # magnitude spectrum of positive frequencies in dB
    X[:hN].real[np.abs(X[:hN].real) < tol] = 0.0  # for phase calculation set to 0 the small values
    X[:hN].imag[np.abs(X[:hN].imag) < tol] = 0.0  # for phase calculation set to 0 the small values
    pX = np.unwrap(np.angle(X[:hN]))  # unwrapped phase spectrum of positive frequencies
    return mX, pX


def dftSynth(mX, pX, M):
    """
	Synthesis of a signal using the discrete Fourier transform
	mX: magnitude spectrum, pX: phase spectrum, M: window size
	returns y: output signal
	"""

    hN = mX.size  # size of positive spectrum, it includes sample 0
    N = (hN - 1) * 2  # FFT size
    if not (UF.isPower2(N)):  # raise error if N not a power of two, thus mX is wrong
        raise ValueError("size of mX is not (N/2)+1")

    hM1 = int(math.floor((M + 1) / 2))  # half analysis window size by rounding
    hM2 = int(math.floor(M / 2))  # half analysis window size by floor
    y = np.zeros(M)  # initialize output array
    Y = np.zeros(N, dtype=complex)  # clean output spectrum
    Y[:hN] = 10 ** (mX / 20) * np.exp(1j * pX)  # generate positive frequencies
    Y[hN:] = 10 ** (mX[-2:0:-1] / 20) * np.exp(-1j * pX[-2:0:-1])  # generate negative frequencies
    fftbuffer = np.real(ifft(Y))  # compute inverse FFT
    y[:hM2] = fftbuffer[-hM2:]  # undo zero-phase window
    y[hM2:] = fftbuffer[:hM1]
    return y



import torch
import torch.nn as nn

from ..utils import * 
from .basic_block import *
from ..filters.analog_filters import FreqRespSampConv1d

class TCNBlock(nn.Module):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        kernel_size: int,
        dilation: int, 
        causal: bool,
        filmed: bool, 
        hypered: bool,
        bias: bool,
    ):
        super().__init__()
        self.causal = causal 
        self.filmed = filmed
        self.hypered = hypered

        if hypered:
            self.conv1 = HyperConv(
                n_inp,
                n_output,
                kernel_size,
                dilation,
                n_cond,
                bias=bias,
            )
        else:
            self.conv1 = nn.Conv1d(n_inp, 
                                    n_output, 
                                    kernel_size = kernel_size, 
                                    padding = 0, # zero padding
                                    dilation = dilation,
                                    bias = bias) # better initialization 

        if filmed and n_cond > 0:
            self.affine = FiLM(n_cond, n_output)
        else:
            self.affine = nn.BatchNorm1d(n_output)
        
        self.activation_function = nn.LeakyReLU(0.2) 
        self.res = nn.Conv1d(n_inp, 
                            n_output, 
                            kernel_size=1,
                            bias=False)
        
    def forward(self, x, c, *args):
        x_in = x

        if self.hypered:
            x = self.conv1(x, c)
        else:
            x = self.conv1(x)

        if c is not None and self.filmed: 
            x = self.affine(x, c)
        else:
            x = self.affine(x)

        x = self.activation_function(x)
        x_res = self.res(x_in)

        if self.causal: 
            x = x + causal_crop(x_res, x.shape[-1])
        else:
            x = x + center_crop(x_res, x.shape[-1])
        return x

class AnalogTCNBlock(nn.Module):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        kernel_size: int,
        dilation: int, 
        causal: bool,
        bias: bool,
        n_samples: int = 640
    ):
        super().__init__()
        self.out_ch = n_output 
        self.causal = causal
        
        self.conv1 = FreqRespSampConv1d(
            in_channels=n_inp,
            out_channels=n_output,
            n_samples=n_samples,
            dilation=dilation,
        )

        self.affine = nn.BatchNorm1d(n_output)

        self.activation_function = nn.LeakyReLU(0.2) 
        self.res = nn.Conv1d(n_inp, 
                            n_output, 
                            kernel_size=1,
                            bias=False)
        
    def forward(self, x, c, *args):
        x_in = x

        x = self.conv1(x)

        x = self.affine(x)

        x = self.activation_function(x)
        x_res = self.res(x_in)

        if self.causal: 
            x = x + causal_crop(x_res, x.shape[-1])
        else:
            x = x + center_crop(x_res, x.shape[-1])
        return x

    def prepare(self, sample_rate, kernel_size, stride=1):
        self.conv1.prepare(sample_rate, kernel_size, stride)

class GCNBlock(nn.Module):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        kernel_size: int,
        dilation: int, 
        causal: bool,
        filmed: bool, 
        hypered: bool,
        bias: bool,
    ):
        super().__init__()
        self.causal = causal 
        self.filmed = filmed
        self.hypered = hypered
        self.out_ch = n_output 

        if hypered:
            self.conv1 = HyperConv(
                n_inp,
                n_output * 2,
                kernel_size,
                dilation,
                n_cond,
                bias=bias,
            )
        else:
            self.conv1 = nn.Conv1d(n_inp, 
                                    n_output * 2, 
                                    kernel_size = kernel_size, 
                                    padding = 0, # zero padding
                                    dilation = dilation,
                                    bias = bias) # better initialization 

        if filmed and n_cond > 0:
            self.affine = FiLM(n_cond, n_output)
        
        
        self.res = nn.Conv1d(n_inp, 
                            n_output, 
                            kernel_size=1,
                            bias=False)
        
        self.mix = torch.nn.Conv1d(n_output,
                                   n_output,
                                   kernel_size = 1,
                                   stride = 1,
                                   padding = 0,
                                   bias = bias)
        
    def forward(self, x, c, *args):
        x_in = x

        if self.hypered:
            x = self.conv1(x, c)
        else:
            x = self.conv1(x)

        # gated mechanism 
        x = torch.tanh(x[:, :self.out_ch, :]) * torch.sigmoid(x[:, self.out_ch:, :])

        if c is not None and self.filmed: 
            x = self.affine(x, c)
        
        z = x 

        x = self.mix(x) 
        x_res = self.res(x_in)

        if self.causal: 
            x = x + causal_crop(x_res, x.shape[-1])
        else:
            x = x + center_crop(x_res, x.shape[-1])
        
        return x, z 
    
class AnalogGCNBlock(nn.Module):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        kernel_size: int,
        dilation: int, 
        causal: bool,
        bias: bool,
        n_samples: int = 640
    ):
        super().__init__()
        self.out_ch = n_output 
        self.causal = causal

        self.conv1 = FreqRespSampConv1d(
            in_channels=n_inp,
            out_channels=n_output * 2,
            n_samples=n_samples,
            dilation=dilation,
        )

        self.res = nn.Conv1d(n_inp, 
                            n_output, 
                            kernel_size=1,
                            bias=False)
        
        self.mix = torch.nn.Conv1d(n_output,
                                   n_output,
                                   kernel_size = 1,
                                   stride = 1,
                                   padding = 0,
                                   bias = bias)
        
    def forward(self, x, c, *args):
        x_in = x

        x = self.conv1(x)

        # gated mechanism 
        x = torch.tanh(x[:, :self.out_ch, :]) * torch.sigmoid(x[:, self.out_ch:, :])

        z = x 

        x = self.mix(x) 
        x_res = self.res(x_in)

        if self.causal: 
            x = x + causal_crop(x_res, x.shape[-1])
        else:
            x = x + center_crop(x_res, x.shape[-1])
        
        return x, z 

    def prepare(self, sample_rate, kernel_size, stride=1):
        self.conv1.prepare(sample_rate, kernel_size, stride)





import torch
import pyloudnorm as pyln
import scipy.signal as signal
import torch.nn.functional as F

def convert_tensor_to_numpy(tensor, is_squeeze=True):
    """Utils functions for converting tensor to numpy 
    """
    if is_squeeze:
        tensor = tensor.squeeze()
    if tensor.requires_grad:
        tensor = tensor.detach()
    if tensor.is_cuda:
        tensor = tensor.cpu()
    return tensor.numpy()

def comlex_to_magnitude(stft_complex):
    """Utils functions for converting complex spectrogram to maginitude 
    """
    stft_mag = torch.sum(stft_complex ** 2) ** 0.5
    return stft_mag

def loudness(x, sample_rate):
    # codes from : https://github.com/adobe-research/DeepAFx-ST/blob/main/deepafx_st/metrics.py
    """Compute the loudness in dB LUFS of waveform."""
    meter = pyln.Meter(sample_rate)

    # add stereo dim if needed
    if x.shape[0] < 2:
        x = x.repeat(2, 1)
    
    return torch.tensor(meter.integrated_loudness(x.permute(1, 0).numpy()))

def crest_factor(x):
    # codes from https://github.com/adobe-research/DeepAFx-ST/blob/main/deepafx_st/metrics.py
    """Compute the crest factor of waveform."""

    peak, _ = x.abs().max(dim=-1)
    rms = torch.sqrt((x ** 2).mean(dim=-1))

    return 20 * torch.log(peak / rms.clamp(1e-8))


def rms_energy(x):
    # codes from https://github.com/adobe-research/DeepAFx-ST/blob/main/deepafx_st/metrics.py
    """Compute the rms energy of waveform
    """
    rms = torch.sqrt((x ** 2).mean(dim=-1))

    return 20 * torch.log(rms.clamp(1e-8))

def spectral_centroid(x):

    spectrum = torch.fft.rfft(x).abs()
    normalized_spectrum = spectrum / spectrum.sum()
    normalized_frequencies = torch.linspace(0, 1, spectrum.shape[-1])
    spectral_centroid = torch.sum(normalized_frequencies * normalized_spectrum)

    return spectral_centroid

# > ================================================== <
#   High pass filter, for better reconstructing high-freq contents, proposed by VV
# > ================================================== <
class DC_PreEmph(torch.nn.Module):
    """
        code from: https://github.com/Alec-Wright/GreyBoxDRC/blob/main/loss_funcs.py
    """
    def __init__(self, R=0.995):
        super().__init__()

        t, ir = signal.dimpulse(signal.dlti([1, -1], [1, -R]), n=2000)
        ir = ir[0][:, 0]

        self.zPad = len(ir) - 1
        self.pars = torch.flipud(torch.tensor(ir, requires_grad=False, dtype=torch.FloatTensor.dtype)).unsqueeze(0).unsqueeze(0)
        
    def forward(
        self, 
        output: torch.tensor, 
        target: torch.tensor
    ):  
        # zero pad the input/target so the filtered signal is the same length
        output = torch.cat((torch.zeros(output.shape[0], 1, self.zPad).type_as(output), output), dim=2)
        target = torch.cat((torch.zeros(output.shape[0], 1, self.zPad).type_as(output), target), dim=2)

        output = torch.nn.functional.conv1d(output, self.pars.type_as(output), bias=None) # [B, 1, T]
        target = torch.nn.functional.conv1d(target, self.pars.type_as(output), bias=None) # [B, 1, T]

        return output, target



class STNSeparation(torch.nn.Module):
    def __init__(
        self, 
        sr,
        n_fft=4096, 
        hop_length=1024, 
        beta_u1=0.8, 
        beta_l1=0.7, 
        beta_u2=0.85, 
        beta_l2=0.75
    ):
        super().__init__()
        self.sr = sr 
        self.n_fft = n_fft
        self.hop_length = hop_length
        self.beta_u1 = beta_u1
        self.beta_l1 = beta_l1
        self.beta_u2 = beta_u2
        self.beta_l2 = beta_l2

        self.L_h = max(3, int(round(0.2 * sr / hop_length)))
        if self.L_h % 2 != 1:
            self.L_h += 1
        self.L_v = max(3, int(round(500 * n_fft / sr)))
        if self.L_v % 2 != 1:
            self.L_v += 1

        self.n_fft_2 = 512
        self.hop_length_2 = 128
        self.L_h_2 = max(3, int(round(0.2 * sr / self.hop_length_2)))
        if self.L_h_2 % 2 != 1:
            self.L_h_2 += 1
        self.L_v_2 = max(3, int(round(500 * self.n_fft_2 / sr)))
        if self.L_v_2 % 2 != 1:
            self.L_v_2 += 1
        

    def forward(self, x):
        # x: [b, 1, len]
        # Step 1: STFT
        
        x = x.squeeze(1)
        X = torch.stft(
            x,
            n_fft=self.n_fft, 
            hop_length=self.hop_length, 
            win_length=self.n_fft, 
            return_complex=True
        )
        X_mag = torch.abs(X)
        
        X_h = self.median_filter_time(X_mag, self.L_h)
        X_v = self.median_filter_freq(X_mag, self.L_v)
        
        R_s = X_h / (X_h + X_v + 1e-8)

        S = self.compute_masks(R_s, self.beta_u1, self.beta_l1)
        R = 1 - S

        X_s = S * X
        X_r = R * X

        x_s = torch.istft(X_s, n_fft=self.n_fft, hop_length=self.hop_length)
        x_r = torch.istft(X_r, n_fft=self.n_fft, hop_length=self.hop_length)

        X_r = torch.stft(
            x_r, 
            n_fft=self.n_fft_2, 
            hop_length=self.hop_length_2, 
            win_length=self.n_fft_2, 
            return_complex=True
        )
        X_abs_2 = torch.abs(X_r)

        X_h_2 = self.median_filter_time(X_abs_2, self.L_h_2)
        X_v_2 = self.median_filter_freq(X_abs_2, self.L_v_2)

        R_t = X_v_2 / (X_h_2 + X_v_2 + 1e-8)
        T = self.compute_masks(R_t, self.beta_u2, self.beta_l2)
        N = 1 - T

        X_t = T * X_r
        X_n = N * X_r


        x_t = torch.istft(X_t, n_fft=self.n_fft_2, hop_length=self.hop_length_2)
        x_n = torch.istft(X_n, n_fft=self.n_fft_2, hop_length=self.hop_length_2)

        # Ensure all outputs have the same length
        min_length = min(x_s.shape[-1], x_t.shape[-1], x_n.shape[-1])
        x_s = x_s[..., :min_length]
        x_t = x_t[..., :min_length]
        x_n = x_n[..., :min_length]

        return x_s, x_t, x_n

    def median_filter_time(self, x, median_h):
        # x shape: (batch, freq, time)
        pad = median_h // 2 
        x_padded = F.pad(x, (pad, pad), mode='reflect')
        x_unfolded = x_padded.unfold(-1, median_h, 1)
        x_filtered = torch.median(x_unfolded, dim=-1)[0]
        return x_filtered[..., :x.size(-1)]

    def median_filter_freq(self, x, median_v):
        # x shape: (batch, freq, time)
        pad = median_v // 2
        x_padded = F.pad(x.transpose(-1, -2), (pad, pad), mode='reflect').transpose(-1, -2)
        x_unfolded = x_padded.unfold(-2, median_v, 1)
        x_filtered = torch.median(x_unfolded, dim=-1)[0]
        return x_filtered[:, :x.size(1), :]

    def compute_masks(self, x, beta_u, beta_l):
        mask = torch.zeros_like(x)
        mask[x >= beta_u] = 1
        transition = (x >= beta_l) & (x < beta_u)
        mask[transition] = torch.sin(torch.pi/2 * (x[transition] - beta_l) / (beta_u - beta_l))**2
        return mask




import torch 
import torch.nn as nn 


class FiLM(nn.Module):
    def __init__(self, cond_dim: int, num_features: int):
        super().__init__()

        self.num_features = num_features
        self.bn = nn.BatchNorm1d(num_features, affine=False)
        self.adaptor = nn.Linear(cond_dim, num_features*2)

    def forward(self, x, cond):
        cond = self.adaptor(cond)
        g, b = torch.chunk(cond, 2, dim=-1)
        
        g = g.unsqueeze(-1)
        b = b.unsqueeze(-1)
        
        x = self.bn(x)      # apply BatchNorm without affine 
        x = (x * g) + b     # then apply conditional affine

        return x
    

class HyperConv(nn.Module):
    def __init__(self,
                n_inp,
                n_output,
                kernel_size,
                dilation,
                cond_size,
                bias=False,):
        super().__init__()
        
        #arguments
        self.in_ch = n_inp
        self.out_ch = n_output
        self.kernel_size = kernel_size
        self.dilation = dilation 
        self.bias = bias

        self.kernel_subnet = nn.Sequential(
            nn.Linear(cond_size, self.in_ch, bias=True),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(self.in_ch, self.in_ch * self.out_ch * kernel_size, bias=True)
        )

        self.bias = bias
        if self.bias:
            self.bias_subnet = nn.Sequential(
                nn.Linear(cond_size, self.in_ch, bias=True),
                nn.LeakyReLU(0.2, inplace=True),
                nn.Linear(self.in_ch, self.out_ch, bias=True)
            )
        

    def forward(self, x, p):
        B = x.shape[0]
        tp = p.unsqueeze(-1)
        padding = self.dilation * (self.kernel_size - 1)
        start, end = padding, x.shape[-1]
        x = torch.cat([
            x[:, :, start-i*self.dilation:end-i*self.dilation] for i in range(self.kernel_size)
        ], dim=1)
        x = x.permute(0, 2, 1).contiguous().view(x.shape[0] * tp.shape[-1], x.shape[-1]//tp.shape[-1], x.shape[1])
        weight = self.kernel_subnet(p).view(B, self.in_ch * self.kernel_size, self.out_ch, tp.shape[-1]) # linear
        weight = weight.permute(0, 3, 1, 2).contiguous().view(B * tp.shape[-1], self.in_ch * self.kernel_size, self.out_ch)
        y = torch.bmm(x, weight)

        if self.bias:
            bias = self.bias_subnet(p).view(B, self.out_ch, tp.shape[-1])
            bias = bias.permute(0, 2, 1).contiguous().view(B * tp.shape[-1], self.out_ch)
            y = y + bias[:, None, :]
        y = y.view(B, -1, self.out_ch).permute(0, 2, 1).contiguous()
        return y


# functions that implement analysis and synthesis of sounds using the Sinusoidal Model
# (for example usage check the examples in models_interface)

import numpy as np
from scipy.signal.windows import blackmanharris, triang
from scipy.fft import ifft, fftshift
import math
from . import dftModel as DFT
from . import utilFunctions as UF
#import utilFunctions as UF

def sineTracking(pfreq, pmag, pphase, tfreq, freqDevOffset=20, freqDevSlope=0.01):
    """
	Tracking sinusoids from one frame to the next
	pfreq, pmag, pphase: frequencies and magnitude of current frame
	tfreq: frequencies of incoming tracks from previous frame
	freqDevOffset: minimum frequency deviation at 0Hz 
	freqDevSlope: slope increase of minimum frequency deviation
	returns tfreqn, tmagn, tphasen: frequency, magnitude and phase of tracks
	"""

    tfreqn = np.zeros(tfreq.size)  # initialize array for output frequencies
    tmagn = np.zeros(tfreq.size)  # initialize array for output magnitudes
    tphasen = np.zeros(tfreq.size)  # initialize array for output phases
    pindexes = np.array(np.nonzero(pfreq), dtype=int)[0]  # indexes of current peaks
    incomingTracks = np.array(np.nonzero(tfreq), dtype=int)[0]  # indexes of incoming tracks
    newTracks = np.zeros(tfreq.size, dtype=int) - 1  # initialize to -1 new tracks
    magOrder = np.argsort(-pmag[pindexes])  # order current peaks by magnitude
    pfreqt = np.copy(pfreq)  # copy current peaks to temporary array
    pmagt = np.copy(pmag)  # copy current peaks to temporary array
    pphaset = np.copy(pphase)  # copy current peaks to temporary array

    # continue incoming tracks
    if incomingTracks.size > 0:  # if incoming tracks exist
        for i in magOrder:  # iterate over current peaks
            if incomingTracks.size == 0:  # break when no more incoming tracks
                break
            track = np.argmin(abs(pfreqt[i] - tfreq[incomingTracks]))  # closest incoming track to peak
            freqDistance = abs(pfreq[i] - tfreq[incomingTracks[track]])  # measure freq distance
            if freqDistance < (freqDevOffset + freqDevSlope * pfreq[i]):  # choose track if distance is small
                newTracks[incomingTracks[track]] = i  # assign peak index to track index
                incomingTracks = np.delete(incomingTracks, track)  # delete index of track in incomming tracks
    indext = np.array(np.nonzero(newTracks != -1), dtype=int)[0]  # indexes of assigned tracks
    if indext.size > 0:
        indexp = newTracks[indext]  # indexes of assigned peaks
        tfreqn[indext] = pfreqt[indexp]  # output freq tracks
        tmagn[indext] = pmagt[indexp]  # output mag tracks
        tphasen[indext] = pphaset[indexp]  # output phase tracks
        pfreqt = np.delete(pfreqt, indexp)  # delete used peaks
        pmagt = np.delete(pmagt, indexp)  # delete used peaks
        pphaset = np.delete(pphaset, indexp)  # delete used peaks

    # create new tracks from non used peaks
    emptyt = np.array(np.nonzero(tfreq == 0), dtype=int)[0]  # indexes of empty incoming tracks
    peaksleft = np.argsort(-pmagt)  # sort left peaks by magnitude
    if ((peaksleft.size > 0) & (emptyt.size >= peaksleft.size)):  # fill empty tracks
        tfreqn[emptyt[:peaksleft.size]] = pfreqt[peaksleft]
        tmagn[emptyt[:peaksleft.size]] = pmagt[peaksleft]
        tphasen[emptyt[:peaksleft.size]] = pphaset[peaksleft]
    elif ((peaksleft.size > 0) & (emptyt.size < peaksleft.size)):  # add more tracks if necessary
        tfreqn[emptyt] = pfreqt[peaksleft[:emptyt.size]]
        tmagn[emptyt] = pmagt[peaksleft[:emptyt.size]]
        tphasen[emptyt] = pphaset[peaksleft[:emptyt.size]]
        tfreqn = np.append(tfreqn, pfreqt[peaksleft[emptyt.size:]])
        tmagn = np.append(tmagn, pmagt[peaksleft[emptyt.size:]])
        tphasen = np.append(tphasen, pphaset[peaksleft[emptyt.size:]])
    return tfreqn, tmagn, tphasen


def cleaningSineTracks(tfreq, minTrackLength=3):
    """
	Delete short fragments of a collection of sinusoidal tracks 
	tfreq: frequency of tracks
	minTrackLength: minimum duration of tracks in number of frames
	returns tfreqn: output frequency of tracks
	"""

    if tfreq.shape[1] == 0:  # if no tracks return input
        return tfreq
    nFrames = tfreq[:, 0].size  # number of frames
    nTracks = tfreq[0, :].size  # number of tracks in a frame
    for t in range(nTracks):  # iterate over all tracks
        trackFreqs = tfreq[:, t]  # frequencies of one track
        trackBegs = np.nonzero((trackFreqs[:nFrames - 1] <= 0)  # begining of track contours
                               & (trackFreqs[1:] > 0))[0] + 1
        if trackFreqs[0] > 0:
            trackBegs = np.insert(trackBegs, 0, 0)
        trackEnds = np.nonzero((trackFreqs[:nFrames - 1] > 0)  # end of track contours
                               & (trackFreqs[1:] <= 0))[0] + 1
        if trackFreqs[nFrames - 1] > 0:
            trackEnds = np.append(trackEnds, nFrames - 1)
        trackLengths = 1 + trackEnds - trackBegs  # lengths of trach contours
        for i, j in zip(trackBegs, trackLengths):  # delete short track contours
            if j <= minTrackLength:
                trackFreqs[i:i + j] = 0
    return tfreq


def sineModel(x, fs, w, N, t):
    """
	Analysis/synthesis of a sound using the sinusoidal model, without sine tracking
	x: input array sound, w: analysis window, N: size of complex spectrum, t: threshold in negative dB 
	returns y: output array sound
	"""

    hM1 = int(math.floor((w.size + 1) / 2))  # half analysis window size by rounding
    hM2 = int(math.floor(w.size / 2))  # half analysis window size by floor
    Ns = 512  # FFT size for synthesis (even)
    H = Ns // 4  # Hop size used for analysis and synthesis
    hNs = Ns // 2  # half of synthesis FFT size
    pin = max(hNs, hM1)  # init sound pointer in middle of anal window
    pend = x.size - max(hNs, hM1)  # last sample to start a frame
    yw = np.zeros(Ns)  # initialize output sound frame
    y = np.zeros(x.size)  # initialize output array
    w = w / sum(w)  # normalize analysis window
    sw = np.zeros(Ns)  # initialize synthesis window
    ow = triang(2 * H)  # triangular window
    sw[hNs - H:hNs + H] = ow  # add triangular window
    bh = blackmanharris(Ns)  # blackmanharris window
    bh = bh / sum(bh)  # normalized blackmanharris window
    sw[hNs - H:hNs + H] = sw[hNs - H:hNs + H] / bh[hNs - H:hNs + H]  # normalized synthesis window
    while pin < pend:  # while input sound pointer is within sound
        # -----analysis-----
        x1 = x[pin - hM1:pin + hM2]  # select frame
        mX, pX = DFT.dftAnal(x1, w, N)  # compute dft
        ploc = UF.peakDetection(mX, t)  # detect locations of peaks
        iploc, ipmag, ipphase = UF.peakInterp(mX, pX, ploc)  # refine peak values by interpolation
        ipfreq = fs * iploc / float(N)  # convert peak locations to Hertz
        # -----synthesis-----
        Y = UF.genSpecSines_p(ipfreq, ipmag, ipphase, Ns, fs)  # generate sines in the spectrum
        fftbuffer = np.real(ifft(Y))  # compute inverse FFT
        yw[:hNs - 1] = fftbuffer[hNs + 1:]  # undo zero-phase window
        yw[hNs - 1:] = fftbuffer[:hNs + 1]
        y[pin - hNs:pin + hNs] += sw * yw  # overlap-add and apply a synthesis window
        pin += H  # advance sound pointer
    return y


def sineModelAnal(x, fs, w, N, H, t, maxnSines=100, minSineDur=.01, freqDevOffset=20, freqDevSlope=0.01):
    """
	Analysis of a sound using the sinusoidal model with sine tracking
	x: input array sound, w: analysis window, N: size of complex spectrum, H: hop-size, t: threshold in negative dB
	maxnSines: maximum number of sines per frame, minSineDur: minimum duration of sines in seconds
	freqDevOffset: minimum frequency deviation at 0Hz, freqDevSlope: slope increase of minimum frequency deviation
	returns xtfreq, xtmag, xtphase: frequencies, magnitudes and phases of sinusoidal tracks
	"""

    if (minSineDur < 0):  # raise error if minSineDur is smaller than 0
        raise ValueError("Minimum duration of sine tracks smaller than 0")

    hM1 = int(math.floor((w.size + 1) / 2))  # half analysis window size by rounding
    hM2 = int(math.floor(w.size / 2))  # half analysis window size by floor
    x = np.append(np.zeros(hM2), x)  # add zeros at beginning to center first window at sample 0
    x = np.append(x, np.zeros(hM2))  # add zeros at the end to analyze last sample
    pin = hM1  # initialize sound pointer in middle of analysis window
    pend = x.size - hM1  # last sample to start a frame
    w = w / sum(w)  # normalize analysis window
    tfreq = np.array([])
    while pin < pend:  # while input sound pointer is within sound
        x1 = x[pin - hM1:pin + hM2]  # select frame
        mX, pX = DFT.dftAnal(x1, w, N)  # compute dft
        ploc = UF.peakDetection(mX, t)  # detect locations of peaks
        iploc, ipmag, ipphase = UF.peakInterp(mX, pX, ploc)  # refine peak values by interpolation
        ipfreq = fs * iploc / float(N)  # convert peak locations to Hertz
        # perform sinusoidal tracking by adding peaks to trajectories
        tfreq, tmag, tphase = sineTracking(ipfreq, ipmag, ipphase, tfreq, freqDevOffset, freqDevSlope)
        tfreq = np.resize(tfreq, min(maxnSines, tfreq.size))  # limit number of tracks to maxnSines
        tmag = np.resize(tmag, min(maxnSines, tmag.size))  # limit number of tracks to maxnSines
        tphase = np.resize(tphase, min(maxnSines, tphase.size))  # limit number of tracks to maxnSines
        jtfreq = np.zeros(maxnSines)  # temporary output array
        jtmag = np.zeros(maxnSines)  # temporary output array
        jtphase = np.zeros(maxnSines)  # temporary output array
        jtfreq[:tfreq.size] = tfreq  # save track frequencies to temporary array
        jtmag[:tmag.size] = tmag  # save track magnitudes to temporary array
        jtphase[:tphase.size] = tphase  # save track magnitudes to temporary array
        if pin == hM1:  # if first frame initialize output sine tracks
            xtfreq = jtfreq
            xtmag = jtmag
            xtphase = jtphase
        else:  # rest of frames append values to sine tracks
            xtfreq = np.vstack((xtfreq, jtfreq))
            xtmag = np.vstack((xtmag, jtmag))
            xtphase = np.vstack((xtphase, jtphase))
        pin += H
    # delete sine tracks shorter than minSineDur
    xtfreq = cleaningSineTracks(xtfreq, round(fs * minSineDur / H))
    return xtfreq, xtmag, xtphase


def sineModelSynth(tfreq, tmag, tphase, N, H, fs):
    """
	Synthesis of a sound using the sinusoidal model
	tfreq,tmag,tphase: frequencies, magnitudes and phases of sinusoids
	N: synthesis FFT size, H: hop size, fs: sampling rate
	returns y: output array sound
	"""

    hN = N // 2  # half of FFT size for synthesis
    L = tfreq.shape[0]  # number of frames
    pout = 0  # initialize output sound pointer
    ysize = H * (L + 3)  # output sound size
    y = np.zeros(ysize)  # initialize output array
    sw = np.zeros(N)  # initialize synthesis window
    ow = triang(2 * H)  # triangular window
    sw[hN - H:hN + H] = ow  # add triangular window
    bh = blackmanharris(N)  # blackmanharris window
    bh = bh / sum(bh)  # normalized blackmanharris window
    sw[hN - H:hN + H] = sw[hN - H:hN + H] / bh[hN - H:hN + H]  # normalized synthesis window
    lastytfreq = tfreq[0, :]  # initialize synthesis frequencies
    ytphase = 2 * np.pi * np.random.rand(tfreq[0, :].size)  # initialize synthesis phases
    for l in range(L):  # iterate over all frames
        if (tphase.size > 0):  # if no phases generate them
            ytphase = tphase[l, :]
        else:
            ytphase += (np.pi * (lastytfreq + tfreq[l, :]) / fs) * H  # propagate phases
        Y = UF.genSpecSines_p(tfreq[l,:], tmag[l,:], ytphase, N, fs)  # generate sines in the spectrum (python version)
        #Y = UF.genSpecSines(tfreq[l, :], tmag[l, :], ytphase, N, fs)  # generate sines in the spectrum
        lastytfreq = tfreq[l, :]  # save frequency for phase propagation
        ytphase = ytphase % (2 * np.pi)  # make phase inside 2*pi
        yw = np.real(fftshift(ifft(Y)))  # compute inverse FFT
        y[pout:pout + N] += sw * yw  # overlap-add and apply a synthesis window
        pout += H  # advance sound pointer
    y = np.delete(y, range(hN))  # delete half of first window
    y = np.delete(y, range(y.size - hN, y.size))  # delete half of the last window
    return y



'''
    This code is highly inspired and referenced by: sms-tools 
    https://github.com/MTG/sms-tools 
'''
import numpy as np 

from .sms_tools import utilFunctions as UF
from .sms_tools import sineModel as SM
from .sms_tools import stochasticModel as STM

from scipy.signal import get_window
import torch_dct as dct
import torch
import time

def get_config():
    window ='hann'
    M = 2048
    N = 2048 
    t = -90
    minSineDur = 0.001
    nH = 60
    minf0 = 60
    maxf0 = 8000
    f0et = 7
    harmDevSlope = 0.01
    stocf = 0.04 #0.04
    
    maxnSines = 256
    freqDevOffset = 20
    freqDevSlope = 0.001
    
    # no need to modify anything after this
    Ns = 512
    H = 128
    return {
        'sms_sm':{
            'window': window,
            'M': M,
            'N': N,
            't': t,
            'minSineDur': minSineDur,
            'maxnSines': maxnSines,
            'freqDevOffset': freqDevOffset,
            'freqDevSlope': freqDevSlope,
            'stocf': stocf,
            'Ns': Ns,
            'H': H
        },
        'tms_transient_sm':{
            'window': window,
            'M': 512,
            'N': 512,
            't': t,
            'minSineDur': 0.00003, 
            'maxnSines': maxnSines,
            'freqDevOffset': freqDevOffset,
            'freqDevSlope': freqDevSlope,
            'Ns': Ns,
            'H': H
        }, 
        'tms_transient_noise':{
            'H': H,
            'stocf': stocf,
        }, 
    }


def transient_synth(transient_signal, fs, config, transient_type, dct_blocksize):
    """Transient synthesis algorithm based on the sinusoidal analysis/synthesis methods on the DCT domain
    """
    transient_results = []
    dct_transient_results = []
    start, end = 0, int(fs * dct_blocksize)
    while start < len(transient_signal):
        transient1 = transient_signal[start:end] 
        transient1_dct = dct.dct(torch.from_numpy(transient1), "ortho").numpy() # turn to dct domain
        # Sin Anal & Synth (DCT domain)
        w = get_window(config[transient_type]['window'], config[transient_type]['M'], fftbins=True)
        tfreq, tmag, tphase = SM.sineModelAnal(
                transient1_dct, 
                fs, 
                w, 
                config[transient_type]['N'], 
                config[transient_type]['H'], 
                config[transient_type]['t'], 
                config[transient_type]['maxnSines'], 
                config[transient_type]['minSineDur'], 
                config[transient_type]['freqDevOffset'],
                config[transient_type]['freqDevSlope'])
        tms_transient1_dct = SM.sineModelSynth(tfreq, tmag, tphase, config[transient_type]['Ns'], config[transient_type]['H'], fs)
        dct_transient_results.append(tms_transient1_dct)
        # IDCT to time domain
        tms_transient1 = dct.idct(torch.from_numpy(tms_transient1_dct), "ortho").numpy()
        transient_results.append(tms_transient1)
        
        # next frame (no overlap)
        start = int(end)
        end = start + int(fs * dct_blocksize)
        if end >= len(transient_signal):
            end = len(transient_signal)
    return np.concatenate(transient_results, 0).astype(np.float32).reshape(-1), np.concatenate(dct_transient_results, 0).astype(np.float32).reshape(-1), transient_results, dct_transient_results


def tms_synthesis_tools(wav, fs):
    harmonic_type = 'sms_sm'
    transient_type = 'tms_transient_sm'
    noise_type = 'tms_transient_noise'
    config = get_config()
    #print('Harmonic ')
    st_time = time.time()
    # harmonic 
    
    w = get_window(config[harmonic_type]['window'], config[harmonic_type]['M'], fftbins=True)
    tfreq, tmag, tphase = SM.sineModelAnal(
        wav, 
        fs, 
        w, 
        config[harmonic_type]['N'], 
        config[harmonic_type]['H'], 
        config[harmonic_type]['t'], 
        config[harmonic_type]['maxnSines'], 
        config[harmonic_type]['minSineDur'], 
        config[harmonic_type]['freqDevOffset'],
        config[harmonic_type]['freqDevSlope'])
    tms_harmonic = SM.sineModelSynth(tfreq, tmag, tphase, config[harmonic_type]['Ns'], config[harmonic_type]['H'], fs)
    ed_time = time.time()
    #print('Harmonics: ', ed_time-st_time)

    #print('Transient')
    st_time = time.time()
    #print('tms_harmonic: ', tms_harmonic.shape)
    # transient 
    min_len = min([len(tms_harmonic), len(wav)])
    tms_first_residual = UF.sineSubtraction(wav, config[harmonic_type]['Ns'], config[harmonic_type]['H'], tfreq, tmag, tphase, fs)  #wav[:min_len] - tms_harmonic[:min_len]
    tms_first_residual = wav[:min_len] - tms_harmonic[:min_len]

    # for long wavs
    dct_blocksize = 2 #sec
    tms_transient, tms_dct_transient, transients, dct_transients = transient_synth(tms_first_residual, fs, config, transient_type, dct_blocksize)
    ed_time = time.time()
    #print('Transients: ', ed_time-st_time)
    return tms_dct_transient






import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np 
from typing import Callable, List, Optional, Tuple, Union, Dict, Any
from .utils import comlex_to_magnitude, DC_PreEmph

#################################################################
######          Base Class of the Loss Function            ######
#################################################################
class _Loss_Base(torch.nn.Module):
    def __init__(self, pre_emp: bool = False):
        super().__init__()

        self.pre_emp_filter = None
        if pre_emp:
            self.pre_emp_filter = DC_PreEmph()


    def forward(self, predict: torch.tensor, target: torch.tensor): 
        pass 

#########################################
######          L1 loss            ######
#########################################
class L1Loss(_Loss_Base):
    def __init__(self, pre_emp: bool = False):
        super().__init__(pre_emp)
        print('> [Loss] --- Temporal L1 Loss ---')

    def forward(
        self, 
        predict: torch.tensor, 
        target: torch.tensor
    ):
        if self.pre_emp_filter:
            predict, target = self.pre_emp_filter(predict, target)
        return F.l1_loss(predict, target)


#########################################
######          L2 loss            ######
#########################################
class L2Loss(_Loss_Base):
    def __init__(self, pre_emp: bool = False):
        super().__init__(pre_emp)
        print('> [Loss] --- Temporal L2 Loss ---')

    def forward(
        self, 
        predict: torch.tensor, 
        target: torch.tensor
    ):
        if self.pre_emp_filter:
            predict, target = self.pre_emp_filter(predict, target)
        return torch.mean((predict - target).pow(2))


#########################################
######           STFT              ######
#########################################
class STFTLoss(_Loss_Base):
    def __init__(
        self, 
        win_len: int = 2048, 
        overlap: float = 0.75, 
        is_complex: bool = True,
        pre_emp: bool = False
    ):
        super().__init__(pre_emp)
        print('> [Loss] --- STFT Complex Loss ---')
        print('> [Loss] is complex:', is_complex)
        print('> [Loss] win_len: {}, overlap: {}'.format(win_len, overlap))

        self.win_len     = win_len
        self.is_complex = is_complex
        self.hop_len     = int((1-overlap)*win_len)

        self.window = nn.Parameter(
                torch.from_numpy(
                    np.hanning(win_len)
                ).float(), requires_grad=False)

    def forward(
        self,
        predict: torch.tensor, 
        target: torch.tensor
    ):
        # (B, C, T)  to (B x C, T)
        if self.pre_emp_filter:
            predict, target = self.pre_emp_filter(predict, target)

        predict = predict.reshape(-1, predict.shape[-1])
        target  = target.reshape( -1, target.shape[-1])

        # stft - predict
        stft_predict = torch.stft(
                predict, 
                n_fft=self.win_len, 
                window=self.window, 
                hop_length=self.hop_len, 
                return_complex=True,
                center=False)

        # stft - target
        stft_target = torch.stft(
                target, 
                n_fft=self.win_len, 
                window=self.window, 
                hop_length=self.hop_len, 
                return_complex=True,
                center=False)

        # loss 
        if self.is_complex:
            loss_final = F.l1_loss(stft_predict, stft_target)
        else:
           stft_predict_mag = comlex_to_magnitude(stft_predict)
           stft_target_mag  = comlex_to_magnitude(stft_target)
           loss_final = F.l1_loss(stft_predict_mag, stft_target_mag)

        return loss_final


#########################################
######  Multi-resolution STFT      ######
#########################################
class MRSTFTCLoss(_Loss_Base):
    def __init__(
        self,
        scales: List = [8192, 2048, 512, 128],
        overlap: float = 0.75, 
        pre_emp: bool = False
    ):
        super().__init__(pre_emp)
        print('> [Loss] --- Multi-resolution STFT Loss ---')
    
        self.scales = scales
        self.overlap = overlap
        self.num_scales = len(self.scales)

        self.windows = nn.ParameterList(
            nn.Parameter(torch.from_numpy(np.hanning(scale)).float(), requires_grad=False) for scale in self.scales
        )

    def forward(
        self,
        predict: torch.tensor, 
        target: torch.tensor
    ):
        if self.pre_emp_filter:
            predict, target = self.pre_emp_filter(predict, target)

        # (B, C, T)  to (B x C, T)
        x      = predict.reshape(-1, predict.shape[-1])
        x_orig = target.reshape( -1, target.shape[-1])

        amp = lambda x: x[:,:,:,0]**2 + x[:,:,:,1]**2

        stfts = []
        for i, scale in enumerate(self.scales):
            cur_fft = torch.stft(x, n_fft=scale, window=self.windows[i], hop_length=int((1-self.overlap)*scale), center=False, return_complex=True)
            stfts.append(cur_fft)

        stfts_orig = []
        for i, scale in enumerate(self.scales):
            cur_fft = torch.stft(x_orig, n_fft=scale, window=self.windows[i], hop_length=int((1-self.overlap)*scale), center=False, return_complex=True)
            stfts_orig.append(cur_fft)

        # Compute loss scale x batch
        lin_loss_final = 0
        #log_loss_final = 0
        for i in range(self.num_scales):
            lin_loss = torch.mean(abs(stfts_orig[i] - stfts[i]))
            #log_loss = torch.mean(abs(torch.log(stfts_orig[i] + 1e-4) - torch.log(stfts[i] + 1e-4)))  

            lin_loss_final += lin_loss
            #log_loss_final += log_loss
        
        lin_loss_final /= self.num_scales
        #log_loss_final /= self.num_scales

        return lin_loss_final #+ log_loss_final

#########################################
######         MRSTFT loss         ######
#########################################
class MRSTFTLoss(_Loss_Base):
    def __init__(
        self,
        scales: List = [2048, 512, 128, 32],
        overlap: float = 0.75, 
        pre_emp: bool = False
    ):
        super().__init__(pre_emp)
        print('> [Loss] --- Multi-resolution STFT Loss ---')
    
        self.scales = scales
        self.overlap = overlap
        self.num_scales = len(self.scales)

        self.windows = nn.ParameterList(
            nn.Parameter(torch.from_numpy(np.hanning(scale)).float(), requires_grad=False) for scale in self.scales
        )

    def forward(
        self,
        predict: torch.tensor, 
        target: torch.tensor
    ):
        if self.pre_emp_filter:
            predict, target = self.pre_emp_filter(predict, target)

        # (B, C, T)  to (B x C, T)
        x      = predict.reshape(-1, predict.shape[-1])
        x_orig = target.reshape( -1, target.shape[-1])

        amp = lambda x: x[:,:,:,0]**2 + x[:,:,:,1]**2

        stfts = []
        for i, scale in enumerate(self.scales):
            cur_fft = torch.stft(x, n_fft=scale, window=self.windows[i], hop_length=int((1-self.overlap)*scale), center=False, return_complex=False)
            stfts.append(amp(cur_fft))

        stfts_orig = []
        for i, scale in enumerate(self.scales):
            cur_fft = torch.stft(x_orig, n_fft=scale, window=self.windows[i], hop_length=int((1-self.overlap)*scale), center=False, return_complex=False)
            stfts_orig.append(amp(cur_fft))

        # Compute loss scale x batch
        lin_loss_final = 0
        log_loss_final = 0
        for i in range(self.num_scales):
            lin_loss = torch.mean(abs(stfts_orig[i] - stfts[i]))
            log_loss = torch.mean(abs(torch.log(stfts_orig[i] + 1e-4) - torch.log(stfts[i] + 1e-4)))  

            lin_loss_final += lin_loss
            log_loss_final += log_loss
        
        lin_loss_final /= self.num_scales
        log_loss_final /= self.num_scales

        return lin_loss_final + log_loss_final


#########################################
######           ESR               ######
#########################################
class ESRLoss(_Loss_Base):
    def __init__(self, pre_emp: bool = False):
        super().__init__(pre_emp)
        print('> [Loss] --- ESR Loss ---')

        self.epsilon = 1e-12
    def forward(
        self, 
        predict: torch.tensor, 
        target: torch.tensor
    ):
        if self.pre_emp_filter:
            predict, target = self.pre_emp_filter(predict, target)

        y_pred = predict
        y = target

        loss = torch.add(y, -y_pred)
        loss = torch.pow(loss, 2)
        loss = torch.mean(loss)
        energy = torch.mean(torch.pow(y, 2)) + self.epsilon
        loss = torch.div(loss, energy)
        return loss


#########################################################
######           Eliminating DC Bias               ######
#########################################################
class DCLoss(_Loss_Base):
    def __init__(self, pre_emp: bool = False):
        super().__init__(pre_emp)
        self.epsilon = 0.00001
        
    def forward(
        self, 
        predict: torch.tensor, 
        target: torch.tensor
    ):
        if self.pre_emp_filter:
            predict, target = self.pre_emp_filter(predict, target)

        loss = torch.pow(torch.add(torch.mean(target, 0), -torch.mean(predict, 0)), 2)
        loss = torch.mean(loss)
        energy = torch.mean(torch.pow(target, 2)) + self.epsilon
        loss = torch.div(loss, energy)
        return loss



# for development easily 
class HybridLoss(torch.nn.Module):
    def __init__(self, pre_emp: bool = False):
        super().__init__()
        print('> [Loss] --- Hybrid Trans Loss ---')

        self.mae = L1Loss(pre_emp)
        self.mrstft = MRSTFTLoss(pre_emp=pre_emp)

    def forward(
        self, 
        predict: torch.tensor, 
        target: torch.tensor
    ):  
        return self.mae(predict, target) + 0.1*self.mrstft(predict, target) 




import sys
import os
import glob
import numpy as np
import soundfile as sf

import torch
import torch.nn as nn
from torch.nn import functional as F

import librosa
import librosa.display

import matplotlib.pyplot as plt
from matplotlib import collections as mc
from matplotlib import ticker
from .utils import * 
from .plotting import * 


def plot_distortion_curve(path_outdir, sr, gain, freq, feed_forward_func = None, *args, **kwargs):

    os.makedirs(path_outdir, exist_ok=True)
    path_song_distortion_curve = os.path.join(path_outdir, f'distortion-curve.png')

    n = np.arange(1000)
    x = gain * np.sin(2 * np.pi * n * freq / sr)
    _subx = x
    
    #nn_model = nn_model.to('cpu')
    x = torch.from_numpy(x).unsqueeze(0).unsqueeze(0).float().to('cpu')
    
    ### forward 
    if feed_forward_func is None:
        y = feed_forward_func(x)
    else:
        y = feed_forward_func(x, *args, **kwargs)
    
    y = y.squeeze(0).squeeze(0).detach().numpy()

    # length alignment
    pre_room = len(_subx) - len(y)
    _subx = _subx[pre_room:]

    plt.plot(_subx, y)
    plt.xlabel('Input Gain')
    plt.ylabel('Output Gain')
    plt.tight_layout()
    plt.savefig(path_song_distortion_curve)
    plt.close()

def plot_harmonic_response(path_outdir, sr, gain, freq, feed_forward_func = None, *args, **kwargs):

    os.makedirs(path_outdir, exist_ok=True)
    path_song_harmonic_response = os.path.join(path_outdir, f'harmonic-response.png')

    num = 1000
    n = np.arange(num)
    x = gain * np.sin(2 * np.pi * n * freq / sr)
    _subx = x
    
    #nn_model = nn_model.to('cpu')
    x = torch.from_numpy(x).unsqueeze(0).unsqueeze(0).float().to('cpu')
    
    ### forward 
    if feed_forward_func is None:
        y = feed_forward_func(x)
    else:
        y = feed_forward_func(x, *args, **kwargs)
    
    y = y.squeeze(0).squeeze(0).cpu().detach().numpy()

    # length alignment
    pre_room = len(_subx) - len(y)
    _subx = _subx[pre_room:]

    f = np.linspace(0, sr//2, num= num//2+1)
    H = normalize(np.fft.rfft(y))

    #plt.figure(figsize=(5, 3))

    plt.semilogx(f, 20 * np.log10(np.abs(H)))
    plt.gca().xaxis.set_major_formatter(ticker.ScalarFormatter())
    #plt.rcParams.update({'font.size': 14})
    
    plt.xlabel('Frequency [Hz]')#, fontsize=14
    plt.ylabel('Magnitude [dB]')#, fontsize=14
    # Increase tick label font size
    #plt.tick_params(axis='both', which='major', labelsize=14)
    plt.tight_layout()
    plt.savefig(path_song_harmonic_response, dpi=600)
    plt.close()

def plot_sine_sweep_response_spec(path_outdir, sr, feed_forward_func = None, *args, **kwargs):

    os.makedirs(path_outdir, exist_ok=True)
    path_song_harmonic_response = os.path.join(path_outdir, f'sine-sweep-response-spec.png')

    num = sr * 5
    n = np.arange(num)
    f0 = 20 
    f1 = 20000
    beta = num / np.log(f1 / f0)
    phase = 2 * np.pi * beta * f0 * (pow(f1 / f0, n / num) - 1.0)
    phi = np.pi / 180
    x = np.cos((phase + phi)/sr)

    #x = gain * np.sin(2 * np.pi * n * freq / sr)
    _subx = x
    
    #nn_model = nn_model.to('cpu')
    x = torch.from_numpy(x).unsqueeze(0).unsqueeze(0).float().to('cpu')
    
    ### forward 
    if feed_forward_func is None:
        y = feed_forward_func(x)
    else:
        y = feed_forward_func(x, *args, **kwargs)
    
    y = y.squeeze(0).squeeze(0).cpu().detach().numpy()

    # length alignment
    pre_room = len(_subx) - len(y)
    _subx = _subx[pre_room:]

    # S = np.abs(librosa.stft(_subx))
    # spec = librosa.amplitude_to_db(S, ref=np.max)
    # librosa.display.specshow(
    #     spec, 
    #     y_axis='linear', 
    #     #x_axis='time', 
    #     sr=sr
    # )

    #plt.figure(figsize=(5, 3))
    
    S = np.abs(librosa.stft(y))
    spec = librosa.amplitude_to_db(S, ref=np.max)
    librosa.display.specshow(
        spec, 
        y_axis='linear', 
        #x_axis='time', 
        sr=sr
    )

    #plt.title('Sine Sweep Response')
    
    #plt.rcParams.update({'font.size': 14})
    plt.tight_layout()
    plt.savefig(path_song_harmonic_response, dpi=600)
    plt.close()
    



import librosa
import numpy as np 
from scipy import ndimage


# > for fuzzy logic transient separation 
def stft(x, n_fft, hop_length, win_length):
    return librosa.stft(x, n_fft=n_fft, hop_length=hop_length, window='hann')

def istft(X, hop_length, win_length):
    return librosa.istft(X, hop_length=hop_length, window='hann')

def medfilt_vertical(abs_stft_buffer, n_median_v):
    kernel = np.zeros((n_median_v, 1))
    kernel[:, 0] = 1
    return ndimage.median_filter(abs_stft_buffer, footprint=kernel, mode='nearest')

def medfilt_horizontal(abs_stft_buffer, n_median_h):
    kernel = np.zeros((1, n_median_h))
    kernel[0, :] = 1
    return ndimage.median_filter(abs_stft_buffer, footprint=kernel, mode='nearest')



# for tms tools 




import torch 
import torch.nn as nn 

from abc import ABC, abstractmethod


class NN_Base(torch.nn.Module):
    def __init__(self):
        super().__init__()
    
    def compute_receptive_field(self):
        raise NotImplementedError
    
    def compute_num_of_params(self):
        return (sum(p.numel() for p in self.parameters()), sum(p.numel() for p in self.parameters() if p.requires_grad))


class CNN_Base(NN_Base):
    def __init__(self, kernel_size, sample_rate, dilation_growth, n_blocks):
        super().__init__()
        self.kernel_size = kernel_size 
        self.sample_rate = sample_rate
        self.dilation_growth = dilation_growth
        self.n_blocks = n_blocks

        
    def compute_receptive_field(self): # in samples 
        rf = self.kernel_size
        for n in range(1, self.n_blocks):
            dilation = self.dilation_growth ** n
            rf = rf + ((self.kernel_size-1) * dilation)
        
        return rf, (rf/self.sample_rate) * 1000 # samples, ms

class RNN_Base(NN_Base):
    def __init__(self, sample_rate):
        super().__init__()
        self.sample_rate = sample_rate

    def compute_receptive_field(self): # in samples 
        return 1, (1/self.sample_rate) * 1000 # samples, ms


from .utils import * 


def compute_masks(R, beta_u, beta_l):
    mask = np.zeros_like(R)
    mask[R >= beta_u] = 1
    transition = np.logical_and(beta_l <= R, R < beta_u)
    mask[transition] = np.sin(np.pi/2 * (R[transition] - beta_l) / (beta_u - beta_l)) ** 2
    return mask

def stn_decomposition(x, sr, n_fft=8192, hop_length=2048, beta_u_1=0.8, beta_l_1=0.7, beta_u_2=0.85, beta_l_2=0.75):
    # Stage 1: Sines vs Residual
    X = stft(x, n_fft, hop_length, n_fft)
    
    L_h = max(3, int(round(0.2 * sr / hop_length)))
    L_v = max(3, int(round(500 * n_fft / sr)))
    
    X_abs = np.abs(X)
    X_h = medfilt_horizontal(X_abs, L_h)
    X_v = medfilt_vertical(X_abs, L_v)
    
    R_s = X_h / (X_h + X_v + 1e-8)
    
    S = compute_masks(R_s, beta_u_1, beta_l_1)
    R = 1 - S
    
    X_s = S * X
    X_r = R * X
    
    x_s = istft(X_s, hop_length, n_fft)
    x_r = istft(X_r, hop_length, n_fft)
    
    # Stage 2: Transients vs Noise
    n_fft_2 = 512 
    hop_length_2 = 128 
    
    X_r = stft(x_r, n_fft_2, hop_length_2, n_fft_2)
    
    L_h_2 = max(3, int(round(0.2 * sr / hop_length_2)))
    L_v_2 = max(3, int(round(500 * n_fft_2 / sr)))
    
    X_abs_2 = np.abs(X_r)
    X_h_2 = medfilt_horizontal(X_abs_2, L_h_2)
    X_v_2 = medfilt_vertical(X_abs_2, L_v_2)
    
    R_t = X_v_2 / (X_h_2 + X_v_2 + 1e-8)
    
    T = compute_masks(R_t, beta_u_2, beta_l_2)
    N = 1 - T
    
    X_t = T * X_r
    X_n = N * X_r
    
    x_t = istft(X_t, hop_length_2, n_fft_2)
    x_n = istft(X_n, hop_length_2, n_fft_2)
    
    # Ensure all outputs have the same length
    min_length = min(len(x_s), len(x_t), len(x_n))
    x_s = x_s[:min_length]
    x_t = x_t[:min_length]
    x_n = x_n[:min_length]
    
    return x_s, x_t, x_n


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np 
from typing import Callable, List, Optional, Tuple, Union, Dict, Any
from .utils import comlex_to_magnitude, DC_PreEmph

from ..transient_sep.tms_tools import tms_synthesis_tools
from .utils import comlex_to_magnitude, convert_tensor_to_numpy, loudness, crest_factor, rms_energy, STNSeparation, spectral_centroid
from ..loss_func.loss_func import (
    L1Loss,
    L2Loss,
    MRSTFTLoss,
    STFTLoss,
    ESRLoss
) 

#####################################################
######           Transient metrics             ######
#####################################################
class TransientPreservation(nn.Module):
    """Calculate the transient reconstruction loss (STFT loss) on the DCT domain 

    The algorithm is largely based on
    @article{verma2000extending,
        title={Extending spectral modeling synthesis with transient modeling synthesis},
        author={Verma, Tony S and Meng, Teresa HY},
        journal={Computer music journal},
        pages={47--59},
        year={2000},
        publisher={JSTOR}
    }

    Parameters
    ----------
    sr : int
        the sample rate of the target audio 
    """
    def __init__(self, sr: int):
        super().__init__()
        print('> [Metrics] --- Transient Loss ---')
        self.sr = sr
        self.loss_func = MRSTFTLoss()
        
    def forward(self, predict: torch.Tensor, target: torch.Tensor):
        """Return transient reconstruction loss on the DCT domain by STFT loss 

        Parameters
        ----------
        predict : torch.Tensor
            predicted audio, shape: [batch_size, num_channel, audio_length]
        target : torch.Tensor
            target audio, shape: [batch_size, num_channel, audio_length]

        Returns
        -------
        torch.Tensor
            error value based on transient reconstruction 
        """
        min_len = min(predict.shape[-1], target.shape[-1])
        predict = predict[..., :min_len]
        target = target[..., :min_len]

        predict = convert_tensor_to_numpy(predict)
        target = convert_tensor_to_numpy(target)

        ### block by block (if no block by block, the algorithm is pretty slow)
        # to-do
        block_size = 4
        if min_len <= int(self.sr * block_size):
            pred_dct_transients = tms_synthesis_tools(predict, self.sr)
            target_dct_transients = tms_synthesis_tools(target, self.sr)
            #print('> pred_dct_transients: ', pred_dct_transients.shape)
        else:
            pred_dct_transients_list = []
            target_dct_transients_list = []
            st= 0 
            while (st + int(self.sr * block_size)) < min_len:
                sub_predict = predict[st:st + int(self.sr * block_size)]
                sub_target = target[st:st + int(self.sr * block_size)]
                sub_pred_dct_transients = tms_synthesis_tools(sub_predict, self.sr)
                sub_target_dct_transients = tms_synthesis_tools(sub_target, self.sr)
                pred_dct_transients_list.append(sub_pred_dct_transients)
                target_dct_transients_list.append(sub_target_dct_transients)
                st += int(self.sr * block_size)
            
            pred_dct_transients = np.concatenate(pred_dct_transients_list, axis=-1)
            target_dct_transients = np.concatenate(target_dct_transients_list, axis=-1)

        #pred_dct_transients = tms_synthesis_tools(predict, self.sr)
        #target_dct_transients = tms_synthesis_tools(target, self.sr)
        
        return self.loss_func(
            torch.from_numpy(pred_dct_transients), 
            torch.from_numpy(target_dct_transients)
        )

class TransientPreservation_v2(nn.Module):
    def __init__(self, sr: int):
        super().__init__()
        print('> [Metrics] --- Transient evaluation based on STN separation---')
        self.sr = sr
        self.stn_separation = STNSeparation(sr=sr)
        self.loss_func = ESRLoss()
        
    def forward(self, predict: torch.Tensor, target: torch.Tensor):
        min_len = min(predict.shape[-1], target.shape[-1])
        predict = predict[..., :min_len]
        target = target[..., :min_len]

        _, predict_t, _ = self.stn_separation(predict)
        _, target_t, _ = self.stn_separation(target)

        return self.loss_func(
            predict_t, 
            target_t
        )


#######################################################
######           Level dynamics                  ###### 
#######################################################
# See: https://github.com/adobe-research/DeepAFx-ST/blob/main/deepafx_st/metrics.py
# > ================================================ < # 
# The following code is largely based on https://github.com/adobe-research/DeepAFx-ST/blob/main/deepafx_st/metrics.py
class LUFS(nn.Module):
    """LUFS: Calculate error based on audio loudness meter 

    This part of code is largely based on: https://github.com/csteinmetz1/pyloudnorm

    @inproceedings{steinmetz2021pyloudnorm,
        title={pyloudnorm: {A} simple yet flexible loudness meter in Python},
        author={Steinmetz, Christian J. and Reiss, Joshua D.},
        booktitle={150th AES Convention},
        year={2021}}
    
    Parameters
    ----------
    sr : int
       the sample rate of the audio 
    """
    def __init__(self,  sr: int = 48000):
        super().__init__()
        print('> [Metrics] --- LUFS ---')

        self.sample_rate = sr 
        
    def forward(self, predict: torch.Tensor, target: torch.Tensor):
        """Return the LUFS error (currently supported mono only)

        Parameters
        ----------
        predict : torch.Tensor
            predicted audio, shape: [batch_size, num_channels, audio_length]
        target : torch.Tensor
            target audio, shape: [batch_size, num_channels, audio_length]

        Returns
        -------
        torch.Tensor
            error value based on loudness
        """
        return torch.nn.functional.l1_loss(
            loudness(predict.view(1, -1), self.sample_rate),
            loudness(target.view(1, -1), self.sample_rate),
        )

class CrestFactor(nn.Module): 
    """Calculate the error based on the crest factor (the peak value to the rms value) in units of dB
    """
    def __init__(self,):
        super().__init__()
        print('> [Metrics] --- Crest Factor (Dynamics range) [dB] ---')

    def forward(self, predict: torch.Tensor, target: torch.Tensor):
        return torch.nn.functional.l1_loss(
            crest_factor(predict),
            crest_factor(target),
        )

class RMSEnergy(nn.Module):
    """Calculate the error based on the rms in units of dB
    """
    def __init__(self,):
        super().__init__()
        print('> [Metrics] --- Rms energy (Dynamics range) [dB] ---')

        
    def forward(self, predict: torch.Tensor, target: torch.Tensor):
        return torch.nn.functional.l1_loss(
            rms_energy(predict),
            rms_energy(target),
        )
    
class SpectralCentroid(nn.Module):
    def __init__(self,):
        super().__init__()
        print('> [Metrics] --- Spectral Centroid ---')

        
    def forward(self, predict: torch.Tensor, target: torch.Tensor):
        return torch.nn.functional.l1_loss(
            spectral_centroid(predict),
            spectral_centroid(target),
        )


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch.nn.modules.utils import _single
import functools
import warnings

def erb_to_hz(x):
    """ 
        Convert ERB to Hz
        Args: 
            x (numpy.ndarray or float): Frequency in ERB scale
        Return: 
            numpy.ndarray or float: Frequency in Hz
    """
    return (np.exp(x/9.265)-1) * 24.7 * 9.265 

def hz_to_erb(x):
    '''Convert Hz to ERB

    Args:
        x (numpy.ndarray or float): Frequency in Hz

    Return:
        numpy.ndarray or float: Frequency in ERB scale
    '''
    return np.log(1+x/(24.7*9.265))*9.265


class ModulatedGaussianFilters(nn.Module):
    """ Modulated Gaussian filters

        The frequency response of this filter is given by

        [
            H(\omega) = e^{-(\omega-\omega_{c})^2/(2\sigma^2)} + e^{-(\omega+\omega_{c})^2/(2\sigma^2)}
        ]

        If one_sided is True, this frequency response is changed as

        [
            H(\omega) = e^{-(\omega-\omega_{c})^2/(2\sigma^2)}.
        ]

    """
    def __init__(self,
                n_filters,
                init_type = 'erb',
                min_bw = 1.0 * 2.0 * np.pi,
                initial_freq_range = [20.0, 48000/2],
                one_sided = False,
                init_sigma = 100.0 * 2.0 * np.pi,
                trainable = True):
        '''
        Args:
            n_filters (int): Number of filters
            init_type (str): Initialization type of center frequencies.
                If "erb", set them from initial_freq_range[0] to initial_freq_range[1] with an equal interval in the ERB scale.
                If "linear", set them from initial_freq_range[0] to initial_freq_range[1] with an equal interval in the linear frequency scale.
            min_bw (float): Minimum bandwidth in radian
            initial_freq_range ([float,float]): Initial frequency ranges in Hz, as tuple of minimum (typically 50) and maximum values (typically, half of Nyquist frequency)
            one_sided (bool): If True, ignore the term in the negative frequency region. If False, the corresponding impulse response is modulated Gaussian window.
            init_sigma (float): Initial value for sigma
            trainable (bool): Whether filter parameters are trainable or not.
        '''
        super().__init__()
        lf, hf = initial_freq_range

        if init_type == 'linear':
            mus = np.linspace(lf, hf, n_filters) * 2.0 * np.pi 
        elif init_type == 'erb':
            erb_mus = np.linspace(hz_to_erb(lf), hz_to_erb(hf), n_filters)
            mus = erb_to_hz(erb_mus) * 2.0 * np.pi
        
        sigma2s = init_sigma**2 * np.ones((n_filters,), dtype='f')
        
        self.min_ln_sigma2s = np.log(min_bw**2)
        self.mus = nn.Parameter(torch.from_numpy(mus).float())
        self._ln_sigma2s = nn.Parameter(torch.from_numpy(
            np.log(sigma2s)).float().clamp(min=self.min_ln_sigma2s)
        )

        self.phase = nn.Parameter(
            torch.zeros((n_filters,), dtype=torch.float)
        )
        self.phase.data.uniform_(0.0, np.pi)
        self.one_sided = one_sided

    @property
    def sigma2s(self):
        return self._ln_sigma2s.clamp(min=self.min_ln_sigma2s).exp()

    def get_frequency_responses(self, omega: torch.Tensor):
        """Sample frequency response at omega

        Args:
            omega (torch.Tensor): Angular frequencies (n_angs)
        
        Return
            tuple[torch.Tensor]: Real and imaginary parts of frequency responses sampled at omega.
        """
        if self.one_sided:
            resp_abs = torch.exp(
                -(omega[None,:] - self.mus[:,None]).pow(2.0)/(2.0*self.sigma2s[:,None])
            )
            resp_r = resp_abs * self.phase.cos()[:,None]
            resp_i = resp_abs * self.phase.sin()[:,None]
        else:
            resp_abs = torch.exp(-(omega[None,:] - self.mus[:,None]).pow(2.0)/(2.0*self.sigma2s[:,None])) # n_filters x n_angfreqs
            resp_abs2 = torch.exp(-(omega[None,:] + self.mus[:,None]).pow(2.0)/(2.0*self.sigma2s[:,None])) # to ensure filters whose impulse responses are real.
            resp_r = resp_abs * self.phase.cos()[:,None] + resp_abs2 * ((-self.phase).cos()[:,None])
            resp_i = resp_abs * self.phase.sin()[:,None] + resp_abs2 * ((-self.phase).sin()[:,None])
        return resp_r, resp_i
    
    def extra_repr(self):
        s = f'n_filters={int(self.mus.shape[0])}, one_sided={self.one_sided}'
        return s.format(**self.__dict__)

    @property
    def device(self):
        return self.mus.device
    

class TDModulatedGaussianFilters(ModulatedGaussianFilters):
    def __init__(self, 
                n_filters, 
                train_sample_rate, 
                init_type="erb", 
                min_bw=1.0*2.0*np.pi, 
                initial_freq_range=[50.0, 48000/2], 
                one_sided=False, 
                init_sigma=100.0*2.0*np.pi, 
                trainable=True):
        '''

        Args:
            n_filters (int): Number of filters
            train_sample_rate (float): Trained sampling frequency
            init_type (str): Initialization type of center frequencies.
                If "erb", set them from initial_freq_range[0] to initial_freq_range[1] with an equal interval in the ERB scale.
                If "linear", set them from initial_freq_range[0] to initial_freq_range[1] with an equal interval in the linear frequency scale.
            min_bw (float): Minimum bandwidth in radian
            initial_freq_range ([float,float]): Initial frequency ranges in Hz, as tuple of minimum (typically 50) and maximum values (typically, half of Nyquist frequency)
            one_sided (bool): If True, ignore the term in the negative frequency region. If False, the corresponding impulse response is modulated Gaussian window.
            init_sigma (float): Initial value for sigma
            trainable (bool): Whether filter parameters are trainable or not.
        '''
        super().__init__(n_filters=n_filters, init_type=init_type, min_bw=min_bw, initial_freq_range=initial_freq_range, one_sided=one_sided, init_sigma=init_sigma, trainable=trainable)
        self.register_buffer('train_sample_rate', torch.tensor(float(train_sample_rate)))

    def get_impulse_responses(self, sample_rate: int, tap_size: int):
        '''Sample impulse responses

        Args:
            sample_rate (int): Target sampling frequency
            tap_size (int): Tap size
        
        Return
            torch.Tensor: Sampled impulse responses (n_filters x tap_size)
        '''
        center_freqs_in_hz = self.mus/(2.0*np.pi)
        # check whether the center frequencies are below Nyquist rate
        if self.train_sample_rate > sample_rate:
            mask = center_freqs_in_hz <= sample_rate/2
        ###
        t = (torch.arange(0.0, tap_size, 1).type_as(center_freqs_in_hz)/sample_rate)
        t = (t - t.mean())[None,:]
        ###
        if self.one_sided:
            raise NotImplementedError
        else:
            c = 2.0*(2.0*np.pi*self.sigma2s[:,None]).sqrt()*(-self.sigma2s[:,None]*(t**2)/2.0).exp()
            filter_coeffs = c*(self.mus[:,None] @ t + self.phase[:,None]).cos() # n_filters x tap_size
        if self.train_sample_rate > sample_rate:
            filter_coeffs = filter_coeffs * mask[:,None]
        return filter_coeffs[:,torch.arange(tap_size-1,-1,-1)]



# common modules
def compute_Hilbert_transforms_of_filters(filters):
    '''Compute the Hilber transforms of the input filters

    Args:
        filters (torch.Tensor): Filters (n_filters x kernel_size)

    Return
        torch.Tensor: Hilbert transforms of the weights (out_channels x in_channels x kernel_size)
    '''
    ft_f = torch.fft.rfft(filters, n=filters.shape[1], dim=1, norm="ortho")
    hft_f = torch.view_as_complex(torch.stack((ft_f.imag, -ft_f.real), axis=-1))
    hft_f = torch.fft.irfft(hft_f, n=filters.shape[1], dim=1, norm="ortho")
    return hft_f.reshape(*(filters.shape))


class FreqRespSampConv1d(torch.nn.Module):
    '''SFI convolutional layer using the frequency domain filter design
    '''
    def __init__(self,
                in_channels,
                out_channels,
                n_samples,
                dilation,
                use_Hilbert_transforms=False):
        super().__init__()

        self.in_channels = in_channels 
        self.out_channels = out_channels
        self.n_samples = n_samples
        self.sample_rate = None
        self.kernel_size = None
        self.stride = None
        self.padding = None
        self.dilation = dilation
        self.use_Hilbert_transforms = use_Hilbert_transforms
        
        ## 
        n_filters = in_channels * out_channels
        if self.use_Hilbert_transforms:
            if n_filters%2 == 1:
                raise ValueError(f'n_filters must be even when using Hilbert transforms of filters [n_filters={n_filters}]')
            n_filters //= 2
        self.n_filters = n_filters

        self._cache = dict()

    def prepare(self, sample_rate, kernel_size, stride, padding=0):
        self.sample_rate = sample_rate
        self.kernel_size = _single(kernel_size)
        self.stride = _single(stride)
        self.padding = _single(padding)
        self.continuous_filters = ModulatedGaussianFilters(
            n_filters=self.n_filters,
            #train_sample_rate=self.sample_rate, 
            init_type="erb", 
            min_bw=1.0*2.0*np.pi, 
            initial_freq_range=[50.0, self.sample_rate/2], 
            one_sided=False, 
            init_sigma=100.0*2.0*np.pi, 
            trainable=True
        )

    def _compute_pinvW(self, device):
        kernel_size = self.kernel_size[0]
        sample_rate = self.sample_rate            
        P = (kernel_size-1)//2 if kernel_size%2 == 1 else kernel_size//2
        M = self.n_samples
        nyquist_rate = sample_rate / 2
        #
        ang_freqs = torch.linspace(0, nyquist_rate*2.0*np.pi, M).float().to(device)
        normalized_ang_freqs = ang_freqs / float(sample_rate)
        if kernel_size%2 == 1:
            seq_P = torch.arange(-P, P+1).float()[None,:].to(device)
            ln_W = -normalized_ang_freqs[:,None]*seq_P # M x 2P+1
        else:
            seq_P = torch.arange(-(P-1), P+1).float()[None,:].to(device)
            ln_W = -normalized_ang_freqs[:,None]*seq_P # M x 2P
        ln_W = ln_W.to(device)
        W = torch.cat((torch.cos(ln_W), torch.sin(ln_W)), dim=0) # 2*M x 2P
        ###
        pinvW = torch.pinverse(W) # 2P x 2M
        pinvW.requires_grad_(False)
        ang_freqs.requires_grad_(False)
        return ang_freqs, pinvW


    def approximate_by_FIR(self, device):
        '''Approximate frequency responses of analog filters with those of digital filters

        Args:
            device (torch.Device): Computation device
        
        Return:
            torch.Tensor: Time-reversed impulse responses of digital filters (n_filters x filter_degree (-P to P))
        '''
        
        cache_tag = (self.sample_rate, self.kernel_size, self.stride)
        if cache_tag in self._cache:
            ang_freqs, pinvW = self._cache[cache_tag]
            ang_freqs = ang_freqs.detach().to(device)
            pinvW = pinvW.detach().to(device)
        else:
            ang_freqs, pinvW = self._compute_pinvW(device)
            self._cache[cache_tag] = (ang_freqs.detach().cpu(), pinvW.detach().cpu())
        
        #cache_tag = (self.sample_rate, self.kernel_size, self.stride)
        #ang_freqs, pinvW = self._cache[cache_tag]
        #ang_freqs = ang_freqs.detach().to(device)
        #pinvW = pinvW.detach().to(device)

        ###
        resp_r, resp_i = self.continuous_filters.get_frequency_responses(ang_freqs) # n_filters x M
        resp = torch.cat((resp_r, resp_i), dim=1) # n_filters x 2M
        ###
        fir_coeffs = (pinvW[None,:,:] @ resp[:,:,None])[:,:,0] # n_filters x 2P
        return fir_coeffs[:,torch.arange(self.kernel_size[0]-1,-1,-1)] # time-reversed impulse response

    def weights(self, filters):
        '''Return weights

        Args:
            filters (torch.Tensor): Filters (n_channels x tap_size)

        Return:
            torch.Tensor: Weights. The shape is in_channel x out_channel x tap_size for an SFI convolutional layer
        '''
        filters = self.approximate_by_FIR(self.continuous_filters.device) # n_filters (or n_filters//2) x kernel_size

        if self.use_Hilbert_transforms:
            filters = torch.cat((filters, compute_Hilbert_transforms_of_filters(filters)), dim=0)
        
        return filters.reshape(self.out_channels, self.in_channels, -1)#.cuda()

    def forward(self, input):
        '''

        Args:
            input (torch.Tensor): Input feature (batch x in_channel x time)
        
        Return:
            torch.Tensor: Output feature (batch x out_channel x time)
        '''
        _weights = self.weights(self.continuous_filters)
        return F.conv1d(input, _weights, None, self.stride, self.padding, _single(self.dilation))




import sys
import os
import glob
import numpy as np
import soundfile as sf

import torch
import torch.nn as nn
from torch.nn import functional as F

import librosa
import librosa.display

import matplotlib.pyplot as plt
from matplotlib import collections as mc


### Get Features
mag = lambda x: x[..., 0]**2 + x[...,1]**2

def get_spectrogram_librosa(wav, win_len=512, hop_len=64):
    D = librosa.amplitude_to_db(np.abs(librosa.stft(wav, n_fft=win_len, hop_length=hop_len, center=False)))
    return D

def get_spectrum_librosa(signal, win_len, hop_len):
    spectrogram = np.abs(librosa.stft(signal, n_fft=win_len, hop_length=hop_len, center=False))
    spectrum = np.mean(spectrogram, axis=-1)
    return spectrum ** 0.5

def compute_loudness(signal, win_len, hop_len):
    def rms(sig):
        return np.sqrt(np.mean(sig ** 2))
    
    def mag2db(sig):
        return 20 * np.log10(sig + 10 ** -4)

    res = []
    len_sig = len(signal)
    for st in range(0, len_sig-win_len, hop_len):
        ed = min(len_sig, st + win_len)
        loudness_val = mag2db(rms(signal[st:ed]))
        res.append(loudness_val)
    return np.array(res)

def get_spectrum(tensor_in, win_len, hop_len, device='cpu'):
    window = torch.from_numpy(np.hanning(win_len)).float().to(device)
    stft = torch.stft(
                tensor_in, 
                window=window,
                hop_length=hop_len,
                n_fft=win_len,
                return_complex=False)
    spectrogram = mag(stft)
    spectrum = torch.mean(spectrogram ** 0.5, dim=-1)
    return spectrum ** 0.5

def get_smoothed_spectrum_diff(
        predict, 
        target, 
        win_size=4096,
        overlap=0.75,
        cutoff_freq_ratio=0.02, 
        weighting=True,
        scale=20,
        device='cpu'):
    
    hop_len = int((1 - overlap) * win_size)
    predict = predict.squeeze()
    target = target.squeeze()

    # compute spectrum
    spec_pred = get_spectrum(
            predict,
            win_len=win_size,
            hop_len=hop_len,
            device=device).squeeze() / win_size
    spec_anno = get_spectrum(
            target, 
            win_len=win_size,
            hop_len=hop_len,
            device=device).squeeze() / win_size
    # difference
    spec_diff = scale * (torch.log10(spec_anno) - torch.log10(spec_pred))
    return spec_diff

def normalize(x):
    return x / np.max(np.abs(x))



import torch 
import torch.nn as nn 
from .base import * 



class SnapShotGRU(SnapShot_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        sample_rate: int, 
        layer_norm: bool = False,
        rnn_bias: bool = True, 
    ):
        super().__init__(
            inp_channel = inp_channel, 
            out_channel = out_channel, 
            rnn_size = rnn_size, 
            cell_type = 'gru', 
            sample_rate = sample_rate, 
            layer_norm = layer_norm, 
            rnn_bias = rnn_bias
        )

class ConcatGRU(Concat_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        sample_rate: int, 
        num_conds: int = 0,
        layer_norm: bool = False,
        rnn_bias: bool = True, 
    ):
        super().__init__(
            inp_channel = inp_channel, 
            out_channel = out_channel, 
            rnn_size = rnn_size, 
            cell_type = 'gru', 
            sample_rate = sample_rate, 
            num_conds = num_conds,
            layer_norm = layer_norm, 
            rnn_bias = rnn_bias
        )

class FiLMGRU(FiLM_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        sample_rate: int, 
        n_mlp_blocks: int = 3,
        mlp_size: int = 8,
        num_conds: int = 0,
        layer_norm: bool = False,
        rnn_bias: bool = True, 
    ):
        super().__init__(
            inp_channel = inp_channel, 
            out_channel = out_channel, 
            rnn_size = rnn_size, 
            cell_type = 'gru', 
            sample_rate = sample_rate, 
            n_mlp_blocks = n_mlp_blocks, 
            mlp_size = mlp_size,
            num_conds = num_conds,
            layer_norm = layer_norm, 
            rnn_bias = rnn_bias,
            filmed = True
        )

class StaticHyperGRU(StaticHyper_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        sample_rate: int, 
        n_mlp_blocks: int = 3,
        mlp_size: int = 8,
        num_conds: int = 0,
        layer_norm: bool = False,
        rnn_bias: bool = True, 
    ):
        super().__init__(
            inp_channel = inp_channel, 
            out_channel = out_channel, 
            rnn_size = rnn_size, 
            cell_type = 'gru', 
            sample_rate = sample_rate, 
            n_mlp_blocks = n_mlp_blocks, 
            mlp_size = mlp_size,
            num_conds = num_conds,
            layer_norm = layer_norm, 
            rnn_bias = rnn_bias
        )

class DynamicHyperGRU(DynamicHyper_Base):
    def __init__(
        self,
        inp_channel: int,
        out_channel: int,
        rnn_size: int,
        sample_rate: int, 
        hyper_rnn_size: int = 8,
        n_z_size: int = 8,
        num_conds: int = 0,
        layer_norm: bool = False,
        rnn_bias: bool = True, 
    ):
        super().__init__(
            inp_channel = inp_channel, 
            out_channel = out_channel, 
            rnn_size = rnn_size, 
            cell_type = 'gru', 
            sample_rate = sample_rate, 
            hyper_rnn_size = hyper_rnn_size, 
            n_z_size = n_z_size,
            num_conds = num_conds,
            layer_norm = layer_norm, 
            rnn_bias = rnn_bias
        )




import math 
import torch 
import torch.nn as nn 
from einops import rearrange
from .constants import RNNMAP




class NonlearnableCell(nn.Module): 
    def __init__(self, hidden_size: int, cell_type: str, layer_norm: bool = False):
        super().__init__()

        if cell_type not in ['vanilla_rnn', 'lstm', 'gru']:
            raise ValueError(f'Invalid cell type: {cell_type}, please specify cell type as vanilla_rnn, lstm, or gru')
        

        # layer norm 
        self.norm_i, self.norm_h, self.norm_c = None, None, None
        if cell_type == 'gru':
            self.norm_i = nn.LayerNorm(RNNMAP[cell_type] * hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
            self.norm_h = nn.LayerNorm(RNNMAP[cell_type] * hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
        elif cell_type == 'lstm':
            self.norm_i = nn.LayerNorm(RNNMAP[cell_type] * hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
            self.norm_h = nn.LayerNorm(RNNMAP[cell_type] * hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
            self.norm_c = nn.LayerNorm(hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
        elif cell_type == 'vanilla_rnn':
            self.norm_i = nn.LayerNorm(hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
            self.norm_h = nn.LayerNorm(hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()

        self.cell_type = cell_type 

    def forward_cell(self, x, h, w_ih, w_hh, b_ih = None, b_hh = None):

        if self.cell_type == 'lstm':
            hx, cx = h 
        else:
            hx = h 
        
        # matrix multiplication 
        ih = torch.bmm(x, w_ih) 
        hh = torch.bmm(hx, w_hh)
        if b_ih is not None and b_hh is not None:
            ih = ih + b_ih
            hh = hh + b_hh

        # normalization or identity 
        ih = self.norm_i(ih)
        hh = self.norm_h(hh)

        # forward single step
        if self.cell_type == 'gru':
            return self.forward_gru_cell(hx, ih, hh)
        elif self.cell_type == 'lstm':
            return self.forward_lstm_cell(cx, ih, hh)
        elif self.cell_type == 'vanilla_rnn':
            return self.forward_vanilla_rnn_cell(ih, hh)
        else:
            raise ValueError(f'Detected invalid cell type: {self.cell_type}, please specify cell type as vanilla_rnn, lstm, or gru')
        
    def forward_gru_cell(self, hx, ih, hh):
        i_r, i_i, i_n = ih.chunk(RNNMAP[self.cell_type], -1)
        h_r, h_i, h_n = hh.chunk(RNNMAP[self.cell_type], -1)

        resetgate = torch.sigmoid(i_r + h_r)
        inputgate = torch.sigmoid(i_i + h_i)
        newgate   = torch.tanh(i_n + (resetgate * h_n))
        hy = newgate + inputgate * (hx - newgate)

        return hy, hy
    
    def forward_lstm_cell(self, cx, ih, hh):
        gates = (ih + hh)
        
        ingate, forgetgate, cellgate, outgate = gates.chunk(RNNMAP[self.cell_type], -1)
        
        ingate = torch.sigmoid(ingate)
        forgetgate = torch.sigmoid(forgetgate)
        cellgate = torch.tanh(cellgate)
        outgate = torch.sigmoid(outgate)

        if self.norm_c is not None:
            cy = self.norm_c((forgetgate * cx) + (ingate * cellgate))

        hy = outgate * torch.tanh(cy)
        
        return hy, (hy, cy)
    
    def forward_vanilla_rnn_cell(self, ih, hh):
        nh = torch.tanh(ih + hh)
        return nh, nh
    
    def forward(self, xs, state, w_ih, w_hh, b_ih, b_hh):
        outputs = []
        
        # match the shape 
        if self.cell_type == 'lstm':
            hx, cx = state 
            hx = rearrange(hx, '1 b h -> b 1 h')
            cx = rearrange(cx, '1 b h -> b 1 h')
            state = (hx, cx)
        else:
            state  = rearrange(state, '1 b h -> b 1 h')

        for i in range(xs.shape[1]):
            out, state = self.forward_cell(
                xs[:, i:i+1, ...],
                state,
                w_ih,
                w_hh,
                b_ih,
                b_hh)
            outputs += [out]
        return torch.cat(outputs, dim=1), state


# >>>>>>>>>>>>>>
class LearnableCell(nn.Module):
    def __init__(self, input_size: int , hidden_size: int, cell_type: str, layer_norm: bool = False, bias: bool = True, filmed: bool = False):
        super().__init__()

        # arg 
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.cell_type = cell_type
        self.layer_norm = layer_norm
        self.bias = bias 
        self.filmed = filmed

        # NN 
        self.x2h = nn.Linear(input_size, RNNMAP[cell_type] * hidden_size, bias=bias)
        self.h2h = nn.Linear(hidden_size, RNNMAP[cell_type] * hidden_size, bias=bias)
        
        # layer norm 
        self.norm_i, self.norm_h, self.norm_c = None, None, None
        if cell_type == 'gru':
            self.norm_i = nn.LayerNorm(RNNMAP[cell_type] * hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
            self.norm_h = nn.LayerNorm(RNNMAP[cell_type] * hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
        elif cell_type == 'lstm':
            self.norm_i = nn.LayerNorm(RNNMAP[cell_type] * hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
            self.norm_h = nn.LayerNorm(RNNMAP[cell_type] * hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
            self.norm_c = nn.LayerNorm(hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
        elif cell_type == 'vanilla_rnn':
            self.norm_i = nn.LayerNorm(hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
            self.norm_h = nn.LayerNorm(hidden_size, elementwise_affine = False) if layer_norm else nn.Identity()
        
        # initialize weights 
        self.initialize_weights()

    def forward(self, xs, state, alpha_i = None, beta_i = None, alpha_h = None, beta_h = None):
        
        outputs = []
        
        # match the shape 
        if self.cell_type == 'lstm':
            hx, cx = state 
            try:
                hx = rearrange(hx, '1 b h -> b 1 h')
                cx = rearrange(cx, '1 b h -> b 1 h')
            except:
                hx = hx
                cx = cx
            state = (hx, cx)
        else:
            try:
                state  = rearrange(state, '1 b h -> b 1 h')
            except:
                state = state 
            
        for i in range(xs.shape[1]):
            out, state = self.forward_cell(
                xs[:, i:i+1, ...],
                state,
                alpha_i, beta_i, alpha_h, beta_h)
            outputs += [out]
        return torch.cat(outputs, dim=1), state
    

    def forward_cell(self, x, h, alpha_i = None, beta_i = None, alpha_h = None, beta_h = None):

        if self.cell_type == 'lstm':
            hx, cx = h 
        else:
            hx = h 
        
        # forward single step
        if self.cell_type == 'gru':
            return self.forward_gru_cell(x, hx, alpha_i, beta_i, alpha_h, beta_h)
        elif self.cell_type == 'lstm':
            return self.forward_lstm_cell(x, (hx, cx), alpha_i, beta_i, alpha_h, beta_h)
        elif self.cell_type == 'vanilla_rnn':
            return self.forward_vanilla_rnn_cell(x, hx, alpha_i, beta_i, alpha_h, beta_h)
        else:
            raise ValueError(f'Detected invalid cell type: {self.cell_type}, please specify cell type as vanilla_rnn, lstm, or gru')
        
    def forward_gru_cell(self, x, h, alpha_i = None, beta_i = None, alpha_h = None, beta_h = None):
        
        ih = self.norm_i(self.x2h(x))
        hh = self.norm_h(self.h2h(h))

        if self.filmed:
            ih = (ih * alpha_i.unsqueeze(1)) + beta_i.unsqueeze(1)
            hh = (hh * alpha_h.unsqueeze(1)) + beta_h.unsqueeze(1)  

        i_r, i_i, i_n = ih.chunk(RNNMAP[self.cell_type], -1)
        h_r, h_i, h_n = hh.chunk(RNNMAP[self.cell_type], -1)

        resetgate = torch.sigmoid(i_r + h_r)
        inputgate = torch.sigmoid(i_i + h_i)
        newgate   = torch.tanh(i_n + (resetgate * h_n))
        hy = newgate + inputgate * (h - newgate)
        
        return hy, hy


    def forward_lstm_cell(self, x, h, alpha_i = None, beta_i = None, alpha_h = None, beta_h = None):
        # hidden state and cell state 
        hx, cx = h 

        if self.filmed:
            gates =  (self.norm_i(self.x2h(x)) * alpha_i.unsqueeze(1) + beta_i.unsqueeze(1)) + (self.norm_h(self.h2h(hx)) * alpha_h.unsqueeze(1) + beta_h.unsqueeze(1))
        else:
            gates =  self.norm_i(self.x2h(x)) + self.norm_h(self.h2h(hx))
        
        ingate, forgetgate, cellgate, outgate = gates.chunk(RNNMAP[self.cell_type], -1)
        
        ingate = torch.sigmoid(ingate)
        forgetgate = torch.sigmoid(forgetgate)
        cellgate = torch.tanh(cellgate)
        outgate = torch.sigmoid(outgate)

        if self.norm_c is not None:
            cy = self.norm_c((forgetgate * cx) + (ingate * cellgate))

        hy = outgate * torch.tanh(cy)
        
        return hy, (hy, cy)

    def forward_vanilla_rnn_cell(self, x, h, alpha_i = None, beta_i = None, alpha_h = None, beta_h = None):
        
        if self.filmed:
            wx = self.norm_i(self.x2h(x)) * alpha_i.unsqueeze(1) + beta_i.unsqueeze(1)
            wh = self.norm_h(self.h2h(h)) * alpha_h.unsqueeze(1) + beta_h.unsqueeze(1)
        else:
            wx = self.norm_i(self.x2h(x))
            wh = self.norm_h(self.h2h(h))
        
        n_h = torch.tanh(wx + wh)

        #n_h = n_h.unsqueeze(1)

        return n_h, n_h
    
    def initialize_weights(self):
        std = 1.0 / math.sqrt(self.hidden_size)
        for w in self.parameters():
            w.data.uniform_(-std, std)


class DynamicHyperCell(nn.Module):
    def __init__(self, input_size: int, hidden_size: int, cell_type: str, hyper_size: int, n_z: int, layer_norm: bool = False, bias: bool = True):
        super().__init__()

        # arg 
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.hyper_size = hyper_size
        self.n_z = n_z
        self.cell_type = cell_type
        self.layer_norm = layer_norm
        self.bias = bias 


        self.hyper_rnn = LearnableCell(
            input_size=input_size + hidden_size, 
            hidden_size=hyper_size, 
            cell_type=cell_type,
            layer_norm=layer_norm, 
            bias=bias)

        # dynamic 
        self.z_h = nn.Linear(hyper_size, RNNMAP[cell_type] * n_z)
        self.z_x = nn.Linear(hyper_size, RNNMAP[cell_type] * n_z)
        self.z_b = nn.Linear(hyper_size, RNNMAP[cell_type] * n_z, bias=False)

        d_h = [nn.Linear(n_z, hidden_size, bias=False) for _ in range(RNNMAP[cell_type])]
        self.d_h = nn.ModuleList(d_h)

        d_x = [nn.Linear(n_z, hidden_size, bias=False) for _ in range(RNNMAP[cell_type])]
        self.d_x = nn.ModuleList(d_x)

        d_b = [nn.Linear(n_z, hidden_size) for _ in range(RNNMAP[cell_type])]
        self.d_b = nn.ModuleList(d_b)

        self.w_h = nn.ParameterList([nn.Parameter(torch.randn(hidden_size, hidden_size)) for _ in range(RNNMAP[cell_type])])
        self.w_x = nn.ParameterList([nn.Parameter(torch.randn(hidden_size, input_size)) for _ in range(RNNMAP[cell_type])])


        if layer_norm:
            self.layer_norm = nn.ModuleList([nn.LayerNorm(hidden_size) for _ in range(RNNMAP[cell_type])])
            self.layer_norm_c = nn.LayerNorm(hidden_size)


    def forward(self, xs, cond, h_c, h_c_hat):
        
        outputs = []
        
        for i in range(xs.shape[1]):
            out, h_c, h_c_hat = self.forward_cell(
                xs[:, i:i+1, ...],
                cond,
                h_c,
                h_c_hat)
            outputs += [out.permute(1, 0, 2)]
        return torch.cat(outputs, dim=1), (h_c, h_c_hat)
    
    
    
    def forward_cell(self, x, cond, h_c, h_c_hat):

        # forward single step
        if self.cell_type == 'gru':
            return self.forward_gru_cell(x, cond, h_c, h_c_hat)
        elif self.cell_type == 'lstm':
            return self.forward_lstm_cell(x, cond, h_c, h_c_hat)
        elif self.cell_type == 'vanilla_rnn':
            return self.forward_vanilla_rnn_cell(x, cond, h_c, h_c_hat)
        else:
            raise ValueError(f'Detected invalid cell type: {self.cell_type}, please specify cell type as vanilla_rnn, lstm, or gru')

    def forward_gru_cell(self, x, cond, h_c, h_c_hat):

        h = h_c
        h_hat = h_c_hat 
        
        
        x_hat = torch.cat((h.squeeze(0), cond), dim=-1).unsqueeze(1)

        
        _, h_hat = self.hyper_rnn(x_hat, h_hat)

        h_hat = h_hat.squeeze(0)
        h = h.squeeze(0)
        x = x.squeeze(1)

        z_h = self.z_h(h_hat).chunk(RNNMAP[self.cell_type], dim=-1)
        z_x = self.z_x(h_hat).chunk(RNNMAP[self.cell_type], dim=-1)
        z_b = self.z_b(h_hat).chunk(RNNMAP[self.cell_type], dim=-1)

        
        rin = []



        for i in range(RNNMAP[self.cell_type]):
            d_h = self.d_h[i](z_h[i]).squeeze(1)
            d_x = self.d_x[i](z_x[i]).squeeze(1)
            if i == RNNMAP[self.cell_type]-1:
                y = torch.tanh(
                    rin[0] * d_h * torch.einsum('ij,bj->bi', self.w_h[i], h) + 
                    d_x * torch.einsum('ij,bj->bi', self.w_x[i], x) + 
                    self.d_b[i](z_b[i]).squeeze(1)
                )
            else:
                y = torch.sigmoid(
                    d_h * torch.einsum('ij,bj->bi', self.w_h[i], h) + d_x * torch.einsum('ij,bj->bi', self.w_x[i], x) + self.d_b[i](z_b[i]).squeeze(1)
                )
            if self.layer_norm:
                rin.append(self.layer_norm[i](y))
            else:
                rin.append(y)

        r, i, n = rin # [b, n]
        h_next = n + i * (h - n)
        h_next = h_next.unsqueeze(0)
        
        return h_next, h_next, h_hat
    
    def forward_lstm_cell(self, x, cond, h_c, h_c_hat):
        
        h, c = h_c
        h_hat, c_hat = h_c_hat 
        
        
        x_hat = torch.cat((h.squeeze(0), cond), dim=-1).unsqueeze(1)

        
        _, (h_hat, c_hat) = self.hyper_rnn(x_hat, (h_hat, c_hat))

        h_hat = h_hat.squeeze(0)
        c_hat = c_hat.squeeze(0)
        h = h.squeeze(0)
        c = c.squeeze(0)
        x = x.squeeze(1)

        z_h = self.z_h(h_hat).chunk(RNNMAP[self.cell_type], dim=-1)
        z_x = self.z_x(h_hat).chunk(RNNMAP[self.cell_type], dim=-1)
        z_b = self.z_b(h_hat).chunk(RNNMAP[self.cell_type], dim=-1)

        
        ifgo = []
        for i in range(RNNMAP[self.cell_type]):
            d_h = self.d_h[i](z_h[i]).squeeze(1)
            d_x = self.d_x[i](z_x[i]).squeeze(1)
            y = d_h * torch.einsum('ij,bj->bi', self.w_h[i], h) + d_x * torch.einsum('ij,bj->bi', self.w_x[i], x) + self.d_b[i](z_b[i]).squeeze(1)
            if self.layer_norm:
                ifgo.append(self.layer_norm[i](y))
            else:
                ifgo.append(y)

        i, f, g, o = ifgo
        c_next = torch.sigmoid(f) * c + torch.sigmoid(i) * torch.tanh(g)

        if self.layer_norm:
            h_next = torch.sigmoid(o) * torch.tanh(self.layer_norm_c(c_next))
        else:
            h_next = torch.sigmoid(o) * torch.tanh(c_next)
        
        c_next = c_next.unsqueeze(0)
        h_next = h_next.unsqueeze(0)
        
        
        return h_next, (h_next, c_next), (h_hat, c_hat)

    def forward_vanilla_rnn_cell(self, x, c, h_c, h_c_hat):
        pass

    
    def initialize_weights(self):
        std = 1.0 / math.sqrt(self.hidden_size)
        for w in self.parameters():
            w.data.uniform_(-std, std)


import torch 
import torch.nn as nn 

from ..base import CNN_Base
from ..utils import * 
from .blocks import GCNBlock, AnalogGCNBlock


class SnapShotGCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True,  
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = 0
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else self.n_inp
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                GCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = 0,
                    kernel_size= kernel_size, 
                    dilation=dilation,
                    causal=causal,
                    filmed= False,
                    hypered = False,
                    bias = bias))


        # output mixing layer
        self.mixing_output = torch.nn.Conv1d(
            in_channels=self.n_channels * n_blocks,
            out_channels=self.n_output,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=False
        )

    def forward(self, x, c, *args):
        
        skips = []
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x, zn = block(x, None)
            skips.append(zn)
        
        z = torch.cat([s[:, :, -x.size(2):] for s in skips], dim=1)

        out = self.mixing_output(z)
        
        return out

class SnapShotAnalogGCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True,  
        n_samples: int = 640
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = 0
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else self.n_inp
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                AnalogGCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = 0,
                    kernel_size= kernel_size, 
                    dilation=dilation,
                    causal=causal,
                    bias = bias,
                    n_samples = n_samples))


        # output mixing layer
        self.mixing_output = torch.nn.Conv1d(
            in_channels=self.n_channels * n_blocks,
            out_channels=self.n_output,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=False
        )

        self.prepare(self.sample_rate, self.kernel_size, 1)

    def forward(self, x, c, *args):
        
        skips = []
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x, zn = block(x, None)
            skips.append(zn)
        
        z = torch.cat([s[:, :, -x.size(2):] for s in skips], dim=1)

        out = self.mixing_output(z)
        
        return out

    def prepare(self, sample_rate, kernel_size, stride):
        for b in self.blocks:
            b.prepare(sample_rate, kernel_size, stride)

class ConcatGCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True,  
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = n_cond
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else (self.n_inp + self.n_cond)
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                GCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = 0,
                    kernel_size= kernel_size, 
                    dilation=dilation,
                    causal=causal,
                    filmed= False,
                    hypered = False,
                    bias = bias))


        # output mixing layer
        self.mixing_output = torch.nn.Conv1d(
            in_channels=self.n_channels * n_blocks,
            out_channels=self.n_output,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=False
        )

    def forward(self, x, c, *args):

        if c is not None and len(c.shape) == 2:
            c = c.unsqueeze(-1)
            c = c.expand(-1, -1, x.shape[-1])

        # cat x and cond
        if self.n_cond > 0:
            x = torch.cat((x, c), dim=1)
        
        skips = []
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x, zn = block(x, None)
            skips.append(zn)
        
        z = torch.cat([s[:, :, -x.size(2):] for s in skips], dim=1)

        out = self.mixing_output(z)
        
        return out
    
class FiLMGCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True,  
        pre_film_size: int = 16,
        pre_film_blocks: int = 3,
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = n_cond
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.pre_film = None
        if self.n_cond > 0:
            pre_film = torch.nn.ModuleList()
            for n in range(pre_film_blocks-1):
                inp = pre_film_size if n > 0 else n_cond
                out = pre_film_size
                pre_film.append(nn.Sequential(
                    nn.Linear(inp, out),
                    nn.LeakyReLU(0.2)
                ))
            pre_film.append(nn.Linear(pre_film_size, pre_film_size))
            self.pre_film = nn.Sequential(*pre_film)

        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else self.n_inp
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                GCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = pre_film_size,
                    kernel_size = kernel_size, 
                    dilation = dilation,
                    causal = causal,
                    filmed = True,
                    hypered = False,
                    bias = bias))

        # output mixing layer
        self.mixing_output = torch.nn.Conv1d(
            in_channels=self.n_channels * n_blocks,
            out_channels=self.n_output,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=False
        )

    def forward(self, x, c, *args):
        
        if self.pre_film is not None and self.n_cond > 0:
            c = self.pre_film(c)
        
        skips = []
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x, zn = block(x, c)
            skips.append(zn)
        
        z = torch.cat([s[:, :, -x.size(2):] for s in skips], dim=1)

        out = self.mixing_output(z)
        
        return out

class HyperGCN(CNN_Base):
    def __init__(
        self,
        n_inp: int, 
        n_output: int, 
        n_cond: int,
        sample_rate: int,
        n_blocks: int = 9, 
        kernel_size: int = 3, 
        dilation_growth: int = 2, 
        n_channels: int = 16, 
        causal: bool = True, 
    ):
        super().__init__(kernel_size, sample_rate, dilation_growth, n_blocks)

        # args 
        self.n_inp = n_inp
        self.n_output = n_output
        self.n_cond = n_cond
        self.n_blocks = n_blocks
        self.kernel_size = kernel_size
        self.dilation_growth = dilation_growth
        self.n_channels = n_channels
        self.causal = causal 
        
        self.sample_rate = sample_rate # can used to prevent samplerate mismatch

        # modules 
        self.blocks = torch.nn.ModuleList()
        for n in range(self.n_blocks):
            in_ch = out_ch if n > 0 else self.n_inp
            out_ch = self.n_channels
            dilation = dilation_growth ** n 
            bias = False if n == 0 else True
            self.blocks.append(
                GCNBlock(
                    in_ch, 
                    out_ch, 
                    n_cond = n_cond,
                    kernel_size = kernel_size, 
                    dilation = dilation,
                    causal = causal,
                    filmed = False,
                    hypered = True,
                    bias = bias))

        # output mixing layer
        self.mixing_output = torch.nn.Conv1d(
            in_channels=self.n_channels * n_blocks,
            out_channels=self.n_output,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=False
        )

    def forward(self, x, c, *args):
        
        skips = []
        # iterate over blocks passing conditioning
        for _, block in enumerate(self.blocks):
            x, zn = block(x, c)
            skips.append(zn)
        
        z = torch.cat([s[:, :, -x.size(2):] for s in skips], dim=1)

        out = self.mixing_output(z)
        
        return out




